# stdlib
import logging
import sys
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import names
import pandas as pd

# relative
from ....logger import traceback_and_raise
from ....util import validate_field
from ...common.message import SyftMessage
from ...common.serde.serialize import _serialize as serialize  # noqa: F401
from ...common.uid import UID
from ...io.address import Address
from ...io.location import Location
from ...io.location.specific import SpecificLocation
from ...io.route import Route
from ...node.common.node_service.network_search.network_search_messages import (
    NetworkSearchMessage,
)
from ...pointer.pointer import Pointer
from ...store.proxy_dataset import ProxyDataset
from ...tensor.autodp.adp_tensor import ADPTensor
from ...tensor.tensor import Tensor
from ..abstract.node import AbstractNodeClient
from ..common.action.exception_action import ExceptionMessage
from ..common.client import Client
from ..common.client_manager.association_api import AssociationRequestAPI
from ..common.client_manager.dataset_api import DatasetRequestAPI
from ..common.client_manager.role_api import RoleRequestAPI
from ..common.client_manager.user_api import UserRequestAPI
from ..common.client_manager.vpn_api import VPNAPI
from ..common.node_service.get_remaining_budget.get_remaining_budget_messages import (
    GetRemainingBudgetMessage,
)
from ..common.node_service.node_setup.node_setup_messages import GetSetUpMessage
from ..common.node_service.object_request.object_request_messages import (
    CreateBudgetRequestMessage,
)
from ..common.node_service.object_transfer.object_transfer_messages import (
    LoadObjectMessage,
)
from ..common.node_service.request_receiver.request_receiver_messages import (
    RequestMessage,
)
from ..common.node_service.simple.obj_exists import DoesObjectExistMessage
from ..common.util import check_send_to_blob_storage
from ..common.util import upload_to_s3_using_presigned
from .enums import PyGridClientEnums
from .enums import RequestAPIFields


class RequestQueueClient(AbstractNodeClient):
    def __init__(self, client: Client) -> None:
        self.client = client
        self.handlers = RequestHandlerQueueClient(client=client)

        self.users = UserRequestAPI(client=self)
        self.roles = RoleRequestAPI(client=self)
        self.association = AssociationRequestAPI(client=self)
        self.datasets = DatasetRequestAPI(client=self)

    @property
    def requests(self) -> List[RequestMessage]:

        # relative
        from ..common.node_service.object_request.object_request_messages import (
            GetAllRequestsMessage,
        )

        msg = GetAllRequestsMessage(
            address=self.client.address, reply_to=self.client.address
        )

        requests = self.client.send_immediate_msg_with_reply(msg=msg).requests  # type: ignore

        for request in requests:
            request.gc_enabled = False
            request.owner_client_if_available = self.client

        return requests

    def get_request_id_from_object_id(self, object_id: UID) -> Optional[UID]:
        for req in self.requests:
            if req.object_id == object_id:
                return req.request_id

        return object_id

    def __getitem__(self, key: Union[str, int]) -> RequestMessage:
        if isinstance(key, str):
            for request in self.requests:
                if key == str(request.id.value):
                    return request
            traceback_and_raise(
                KeyError("No such request found for string id:" + str(key))
            )
        if isinstance(key, int):
            return self.requests[key]
        else:
            traceback_and_raise(KeyError("Please pass in a string or int key"))

        raise Exception("should not get here")

    def __repr__(self) -> str:
        return repr(self.requests)

    def _repr_html_(self) -> str:
        return self.pandas._repr_html_()

    @property
    def pandas(self) -> pd.DataFrame:
        # TODO:
        # Replace all the hardcoded string by enums / abstractions.
        request_lines = [
            {
                "Name": req.user_name,
                "Email": req.user_email,
                "Role": req.user_role,
                "Request Type": req.request_type.upper(),  # type: ignore
                "Status": req.status,
                "Reason": req.request_description,
                "Request ID": req.id,
                "Requested Object's ID": req.object_id
                if req.request_type == "data"
                else None,
                "Requested Object's tags": req.object_tags,
                "Requested Budget": req.requested_budget
                if req.request_type == "budget"
                else None,
                "Current Budget": req.current_budget
                if req.request_type == "budget"
                else None,
            }
            for req in self.requests
        ]
        return pd.DataFrame(request_lines)

    def add_handler(
        self,
        action: str,
        print_local: bool = False,
        log_local: bool = False,
        tags: Optional[List[str]] = None,
        timeout_secs: int = -1,
        element_quota: Optional[int] = None,
    ) -> None:
        handler_opts = self._validate_options(
            id=UID(),
            action=action,
            print_local=print_local,
            log_local=log_local,
            tags=tags,
            timeout_secs=timeout_secs,
            element_quota=element_quota,
        )

        self._update_handler(handler_opts, keep=True)

    def remove_handler(self, key: Union[str, int]) -> None:
        handler_opts = self.handlers[key]

        self._update_handler(handler_opts, keep=False)

    def clear_handlers(self) -> None:
        for handler in self.handlers.handlers:
            id_str = str(handler["id"].value).replace("-", "")
            self.remove_handler(id_str)

    def _validate_options(
        self,
        action: str,
        print_local: bool = False,
        log_local: bool = False,
        tags: Optional[List[str]] = None,
        timeout_secs: int = -1,
        element_quota: Optional[int] = None,
        id: Optional[UID] = None,
    ) -> Dict[str, Any]:
        handler_opts: Dict[str, Any] = {}
        if action not in ["accept", "deny"]:
            traceback_and_raise(Exception("Action must be 'accept' or 'deny'"))
        handler_opts["action"] = action
        handler_opts["print_local"] = bool(print_local)
        handler_opts["log_local"] = bool(log_local)

        handler_opts["tags"] = []
        if tags is not None:
            for tag in tags:
                handler_opts["tags"].append(tag)
        handler_opts["timeout_secs"] = max(-1, int(timeout_secs))
        if element_quota is not None:
            handler_opts["element_quota"] = max(0, int(element_quota))

        if id is None:
            id = UID()
        handler_opts["id"] = id

        return handler_opts

    def _update_handler(self, request_handler: Dict[str, Any], keep: bool) -> None:
        # relative
        from ..common.node_service.request_handler.request_handler_messages import (
            UpdateRequestHandlerMessage,
        )

        msg = UpdateRequestHandlerMessage(
            address=self.client.address, handler=request_handler, keep=keep
        )
        self.client.send_immediate_msg_without_reply(msg=msg)


class RequestHandlerQueueClient:
    def __init__(self, client: Client) -> None:
        self.client = client

    @property
    def handlers(self) -> List[Dict]:
        # relative
        from ..common.node_service.request_handler.request_handler_messages import (
            GetAllRequestHandlersMessage,
        )

        msg = GetAllRequestHandlersMessage(
            address=self.client.address, reply_to=self.client.address
        )
        return validate_field(
            self.client.send_immediate_msg_with_reply(msg=msg), "handlers"
        )

    def __getitem__(self, key: Union[str, int]) -> Dict:
        """
        allow three ways to get an request handler:
            1. use id: str
            2. use tag: str
            3. use index: int
        """
        if isinstance(key, str):
            matches = 0
            match_handler: Optional[Dict] = None
            for handler in self.handlers:
                if key in str(handler["id"].value).replace("-", ""):
                    return handler
                if key in handler["tags"]:
                    matches += 1
                    match_handler = handler
            if matches == 1 and match_handler is not None:
                return match_handler
            elif matches > 1:
                raise KeyError("More than one item with tag:" + str(key))

            raise KeyError("No such request found for string id:" + str(key))
        if isinstance(key, int):
            return self.handlers[key]
        else:
            raise KeyError("Please pass in a string or int key")

    def __repr__(self) -> str:
        return repr(self.handlers)

    @property
    def pandas(self) -> pd.DataFrame:
        def _get_time_remaining(handler: dict) -> int:
            timeout_secs = handler.get("timeout_secs", -1)
            if timeout_secs == -1:
                return -1
            else:
                created_time = handler.get("created_time", 0)
                rem = timeout_secs - (time.time() - created_time)
                return round(rem)

        handler_lines = [
            {
                "tags": handler["tags"],
                "ID": handler["id"],
                "action": handler["action"],
                "remaining time (s):": _get_time_remaining(handler),
            }
            for handler in self.handlers
        ]
        return pd.DataFrame(handler_lines)


class DomainClient(Client):

    domain: SpecificLocation
    requests: RequestQueueClient

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        domain: SpecificLocation,
        network: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        version: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            routes=routes,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            version=version,
        )

        self.requests = RequestQueueClient(client=self)

        self.post_init()

        self.users = UserRequestAPI(client=self)
        self.roles = RoleRequestAPI(client=self)
        self.association = AssociationRequestAPI(client=self)
        self.datasets = DatasetRequestAPI(client=self)
        self.vpn = VPNAPI(client=self)

    def obj_exists(self, obj_id: UID) -> bool:
        msg = DoesObjectExistMessage(obj_id=obj_id)
        return self.send_immediate_msg_with_reply(msg=msg).payload  # type: ignore

    @property
    def privacy_budget(self) -> float:
        msg = GetRemainingBudgetMessage(address=self.address, reply_to=self.address)
        return self.send_immediate_msg_with_reply(msg).budget  # type: ignore

    def request_budget(
        self,
        eps: float = 0.0,
        reason: str = "",
        skip_checks: bool = False,
    ) -> Any:

        if not skip_checks:
            if eps == 0.0:
                eps = float(input("Please specify how much more epsilon you want:"))

            if reason == "":
                reason = str(
                    input("Why should the domain owner give you more epsilon:")
                )

        msg = CreateBudgetRequestMessage(
            reason=reason,
            budget=eps,
            address=self.address,
        )

        self.send_immediate_msg_without_reply(msg=msg)

        print(
            "Requested "
            + str(eps)
            + " epsilon of budget. Call .privacy_budget to see if your budget has arrived!"
        )

    def load(
        self, obj_ptr: Type[Pointer], address: Address, pointable: bool = False
    ) -> None:
        content = {
            RequestAPIFields.ADDRESS: serialize(address)
            .SerializeToString()  # type: ignore
            .decode(PyGridClientEnums.ENCODING),
            RequestAPIFields.UID: str(obj_ptr.id_at_location.value),
            RequestAPIFields.POINTABLE: pointable,
        }
        self._perform_grid_request(grid_msg=LoadObjectMessage, content=content)

    def setup(self, *, domain_name: Optional[str], **kwargs: Any) -> Any:
        if domain_name is None:
            domain_name = names.get_full_name() + "'s Domain"
            logging.info(
                "No Domain Name provided... picking randomly as: " + domain_name
            )

        kwargs["domain_name"] = domain_name

        response = self.conn.setup(**kwargs)  # type: ignore
        logging.info(response[RequestAPIFields.MESSAGE])

    def reset(self) -> None:
        logging.warning(
            "Node reset will delete the data, as well as the requests. Do you want to continue (y/N)?"
        )
        response = input().lower()
        if response == "y":
            response = self.routes[0].connection.reset()  # type: ignore

    def search(self, query: List, pandas: bool = False) -> Any:
        response = self._perform_grid_request(
            grid_msg=NetworkSearchMessage, content={RequestAPIFields.QUERY: query}
        )
        if pandas:
            response = pd.DataFrame(response)

        return response

    def _perform_grid_request(
        self, grid_msg: Any, content: Optional[Dict[Any, Any]] = None
    ) -> SyftMessage:
        if content is None:
            content = {}
        # Build Syft Message
        content[RequestAPIFields.ADDRESS] = self.address
        content[RequestAPIFields.REPLY_TO] = self.address
        signed_msg = grid_msg(**content).sign(signing_key=self.signing_key)
        # Send to the dest
        response = self.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def get_setup(self, **kwargs: Any) -> Any:
        return self._perform_grid_request(grid_msg=GetSetUpMessage, content=kwargs)

    def apply_to_network(
        self,
        client: Optional[AbstractNodeClient] = None,
        **metadata: str,
    ) -> None:
        try:
            # joining the network might take some time and won't block

            # Check if the version of the network client is same as the domain client
            if client and hasattr(client, "version"):
                if self.version != client.version:  # type: ignore
                    print(
                        "\n**Warning**: The syft version on your domain and the network are different."
                    )
                    print(
                        f"Domain version: {self.version}\nNetwork Version: {client.version}"  # type: ignore
                    )
            self.join_network(client=client)

            timeout = 30
            connected = False
            network_vpn_ip = ""
            domain_vpn_ip = ""

            # get the vpn ips
            print("Waiting to connect to VPN.")
            while timeout > 0 and connected is False:
                timeout -= 1
                try:
                    vpn_status = self.vpn_status()
                    if vpn_status["connected"]:
                        print("Connected to VPN")
                        connected = True
                        continue
                except Exception as e:
                    print(f"Failed to get vpn status. {e}")
                print(".", end="")
                time.sleep(1)

            for peer in vpn_status["peers"]:
                if peer["hostname"] == client.name:  # type: ignore
                    network_vpn_ip = peer["ip"]
            try:
                domain_vpn_ip = self.vpn_status()["host"]["ip"]
            except Exception as e:
                print(f"Failed to get vpn host ip. {e}")

            if network_vpn_ip == "":
                raise Exception(
                    f"Cant find the network node {client.name} in {vpn_status}"  # type: ignore
                )
            if domain_vpn_ip == "":
                raise Exception(f"No host ip in {vpn_status}")

            self.association.create(
                source=domain_vpn_ip, target=network_vpn_ip, metadata=metadata
            )

            print("Application submitted.")
        except Exception as e:
            print(f"Failed to apply to network with {client}. {e}")

    @property
    def id(self) -> UID:
        return self.domain.id

    @property
    def device(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself, this property will return the Location of that device
        if it is known by the client."""

        return super().device

    @device.setter
    def device(self, new_device: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself and we learn the Location of that device, this setter
        allows us to save the Location of that device for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a domain, you don't need a Device ID.")
        )

    @property
    def vm(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself, this property will return the Location of that vm
        if it is known by the client."""

        return super().vm

    @vm.setter
    def vm(self, new_vm: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself and we learn the Location of that vm, this setter
        allows us to save the Location of that vm for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a device, you don't need a VM Location.")
        )

    def __repr__(self) -> str:
        no_dash = str(self.id).replace("-", "")
        return f"<{type(self).__name__} - {self.name}: {no_dash}>"

    def update_vars(self, state: dict) -> pd.DataFrame:
        for ptr in self.store.store:
            tags = getattr(ptr, "tags", None)
            if tags is not None:
                for tag in tags:
                    state[tag] = ptr
        return self.store.pandas

    def vpn_status(self) -> Dict[str, Any]:
        return self.vpn.get_status()

    def load_dataset(
        self,
        assets: Optional[dict] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        skip_checks: bool = False,
        chunk_size: int = 536870912,  # 500 MB
        use_blob_storage: bool = False,
        **metadata: Dict,
    ) -> None:
        sys.stdout.write("Loading dataset...")
        if assets is None or not isinstance(assets, dict):
            raise Exception(
                "Missing Assets: Oops!... You forgot to include your data! (or you passed it in the wrong way) \n\n"
                "You must call load_dataset() with a dictionary of assets which are the "
                "private dataset objects (tensors) you wish to allow someone else to study "
                "while PySyft protects it using various privacy enhancing technologies. \n\n"
                "For example, the MNIST dataset is comprised of 6 tensors, so we would create an assets "
                "dictionary with 6 keys (strings) mapping to the 6 tensors of MNIST.\n\n"
                "Please pass in a dictionary where the key is the name of the asset and the value is "
                "the private dataset object (tensor) itself. We recommend uploading assets which "
                "are differential-privacy trackable objects, such as a syft.Tensor() wrapped "
                "numpy.int32 or numpy.float32 object which you then call .private() on. \n\nOnce "
                "you have an assets dictionary call load_dataset(assets=<your dict of objects>)."
            )
        sys.stdout.write("\rLoading dataset... checking assets...")

        if name is None:
            raise Exception(
                "Missing Name: Oops!... You forgot to name your dataset!\n\n"
                "It's important to give your dataset a clear and descriptive name because"
                " the name is the primary way in which potential users of the dataset will"
                " identify it.\n\n"
                'Retry with a string name. I.e., .load_dataset(name="<your name here>)"'
            )
        sys.stdout.write("\rLoading dataset... checking dataset name for uniqueness...")

        # Disabling this for now until we have a more efficient means of querying dataset metadata.
        # TODO: enforce name uniqueness through more efficient means.
        # datasets = self.datasets
        # if not skip_checks:
        #     for i in range(len(datasets)):
        #         d = datasets[i]
        #         sys.stdout.write(".")
        #         if name == d.name:
        #             print(
        #                 "\n\nWARNING - Dataset Name Conflict: A dataset named '"
        #                 + name
        #                 + "' already exists.\n"
        #             )
        #             pref = input("Do you want to upload this dataset anyway? (y/n)")
        #             while pref != "y" and pref != "n":
        #                 pref = input(
        #                     "Invalid input '" + pref + "', please specify 'y' or 'n'."
        #                 )
        #             if pref == "n":
        #                 raise Exception("Dataset loading cancelled.")
        #             else:
        #                 print()  # just for the newline
        #                 break

        sys.stdout.write(
            "\rLoading dataset... checking dataset name for uniqueness..."
            "                                                          "
            "                                                          "
        )

        if description is None:
            raise Exception(
                "Missing Description: Oops!... You forgot to describe your dataset!\n\n"
                "It's *very* important to give your dataset a very clear and complete description"
                " because your users will need to be able to find this dataset (the description is used for search)"
                " AND they will need enough information to be able to know that the dataset is what they're"
                " looking for AND how to use it.\n\n"
                "Start by describing where the dataset came from, how it was collected, and how its formatted."
                "Refer to each object in 'assets' individually so that your users will know which is which. Don't"
                " be afraid to be longwinded. :) Your users will thank you."
            )

        sys.stdout.write(
            "\rLoading dataset... checking asset types...                              "
        )

        if not skip_checks:
            for _, asset in assets.items():

                if not isinstance(asset, Tensor) or not isinstance(
                    getattr(asset, "child", None), ADPTensor
                ):
                    raise Exception(
                        "ERROR: all private assets must be NumPy ndarray.int32 assets "
                        + "with proper Differential Privacy metadata applied.\n"
                        + "\n"
                        + "Example: syft.Tensor(np.ndarray([1,2,3,4]).astype(np.int32)).private()\n\n"
                        + "or\n\n"
                        + "Example: syft.Tensor([1,2,3,4]).private()\n\n"
                        + "and then follow the wizard. 🧙"
                    )
                    # print(
                    #     "\n\nWARNING - Non-DP Asset: You just passed in a asset '"
                    #     + asset_name
                    #     + "' which cannot be tracked with differential privacy because it is a "
                    #     + str(type(asset))
                    #     + " object.\n\n"
                    #     + "This means you'll need to manually approve any requests which "
                    #     + "leverage this data. If this is ok with you, proceed. If you'd like to use "
                    #     + "automatic differential privacy budgeting, please pass in a DP-compatible tensor type "
                    #     + "such as by calling .private() on a sy.Tensor with a np.int32 or np.float32 inside."
                    # )
                    #
                    # pref = input("Are you sure you want to proceed? (y/n)")
                    #
                    # while pref != "y" and pref != "n":
                    #     pref = input(
                    #         "Invalid input '" + pref + "', please specify 'y' or 'n'."
                    #     )
                    # if pref == "n":
                    #     raise Exception("Dataset loading cancelled.")

        # serialize metadata
        metadata["name"] = bytes(name, "utf-8")  # type: ignore
        metadata["description"] = bytes(description, "utf-8")  # type: ignore

        for k, v in metadata.items():
            if isinstance(v, str):  # type: ignore
                metadata[k] = bytes(v, "utf-8")  # type: ignore

        # If one of the assets needs to be send to blob_storage, then store all other
        # assets to blob storage as well
        # TODO: Determine use_blob_storage from the clients node metadata
        send_assets_to_blob_storage = any(
            [
                check_send_to_blob_storage(obj=asset, use_blob_storage=use_blob_storage)
                for asset in assets.values()
            ]
        )

        if send_assets_to_blob_storage:
            # upload to blob storage
            proxy_assets: Dict[str, ProxyDataset] = {}
            # send each asset to blob storage and pack the results back
            for asset_name, asset in assets.items():
                proxy_obj = upload_to_s3_using_presigned(
                    client=self,
                    data=asset,
                    chunk_size=chunk_size,
                    asset_name=asset_name,
                    dataset_name=name,
                )
                proxy_assets[asset_name] = proxy_obj

            dataset_bytes = serialize(proxy_assets, to_bytes=True)
        else:
            # upload directly
            dataset_bytes = serialize(assets, to_bytes=True)

        sys.stdout.write("\rLoading dataset... uploading...                        ")
        self.datasets.create_syft(
            dataset=dataset_bytes, metadata=metadata, platform="syft"
        )
        sys.stdout.write(
            "\rLoading dataset... uploading... SUCCESS!                        "
        )

        print(
            "\n\nRun <your client variable>.datasets to see your new dataset loaded into your machine!"
        )

    def create_dataset(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        skip_checks: bool = False,
        **metadata: Dict,
    ) -> None:
        # relative
        from ....lib.python.util import downcast

        if name is None:
            raise Exception(
                "Missing Name: Oops!... You forgot to name your dataset!\n\n"
                "It's important to give your dataset a clear and descriptive name because"
                " the name is the primary way in which potential users of the dataset will"
                " identify it.\n\n"
                'Retry with a string name. I.e., .load_dataset(name="<your name here>)"'
            )

        datasets = self.datasets

        if not skip_checks:
            for i in range(len(datasets)):
                d = datasets[i]
                sys.stdout.write(".")
                if name == d.name:
                    print(
                        "\n\nWARNING - Dataset Name Conflict: A dataset named '"
                        + name
                        + "' already exists.\n"
                    )
                    pref = input("Do you want to upload this dataset anyway? (y/n)")
                    while pref != "y" and pref != "n":
                        pref = input(
                            "Invalid input '" + pref + "', please specify 'y' or 'n'."
                        )
                    if pref == "n":
                        raise Exception("Dataset loading cancelled.")
                    else:
                        print()  # just for the newline
                        break

        if description is None:
            raise Exception(
                "Missing Description: Oops!... You forgot to describe your dataset!\n\n"
                "It's *very* important to give your dataset a very clear and complete description"
                " because your users will need to be able to find this dataset (the description is used for search)"
                " AND they will need enough information to be able to know that the dataset is what they're"
                " looking for AND how to use it.\n\n"
                "Start by describing where the dataset came from, how it was collected, and how its formatted."
                "Refer to each object in 'assets' individually so that your users will know which is which. Don't"
                " be afraid to be longwinded. :) Your users will thank you."
            )

        metadata["name"] = bytes(name, "utf-8")  # type: ignore
        metadata["description"] = bytes(description, "utf-8")  # type: ignore

        for k, v in metadata.items():
            if isinstance(v, str):  # type: ignore
                metadata[k] = bytes(v, "utf-8")  # type: ignore

        assets = downcast({})
        binary_dataset = serialize(assets, to_bytes=True)

        metadata = downcast(metadata)

        self.datasets.create_syft(
            dataset=binary_dataset, metadata=metadata, platform="syft"
        )
        sys.stdout.write("Creating an empty dataset... Creating... SUCCESS!")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/device.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/torch/device.proto\x12\x0esyft.lib.torch"%\n\x06\x44\x65vice\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x05\x62\x06proto3'
)


_DEVICE = DESCRIPTOR.message_types_by_name["Device"]
Device = _reflection.GeneratedProtocolMessageType(
    "Device",
    (_message.Message,),
    {
        "DESCRIPTOR": _DEVICE,
        "__module__": "proto.lib.torch.device_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.Device)
    },
)
_sym_db.RegisterMessage(Device)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _DEVICE._serialized_start = 48
    _DEVICE._serialized_end = 85
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ....proto.core.common.recursive_serde_pb2 import (
    RecursiveSerde as RecursiveSerde_PB,
)
from ....util import get_fully_qualified_name
from ....util import index_syft_by_module_name


def rs_object2proto(self: Any) -> RecursiveSerde_PB:
    # if __attr_allowlist__ then only include attrs from that list
    msg = RecursiveSerde_PB(fully_qualified_name=get_fully_qualified_name(self))

    if self.__attr_allowlist__ is None:
        attribute_dict = self.__dict__.keys()
    else:
        attribute_dict = self.__attr_allowlist__

    for attr_name in attribute_dict:
        if hasattr(self, attr_name):
            msg.fields_name.append(attr_name)
            transforms = self.__serde_overrides__.get(attr_name, None)
            if transforms is None:
                field_obj = getattr(self, attr_name)
            else:
                field_obj = transforms[0](getattr(self, attr_name))
            msg.fields_data.append(sy.serialize(field_obj, to_bytes=True))
    return msg


def rs_proto2object(proto: RecursiveSerde_PB) -> Any:
    class_type = index_syft_by_module_name(proto.fully_qualified_name)
    obj = class_type.__new__(class_type)  # type: ignore
    for attr_name, attr_bytes in zip(proto.fields_name, proto.fields_data):
        attr_value = sy.deserialize(attr_bytes, from_bytes=True)
        transforms = obj.__serde_overrides__.get(attr_name, None)
        try:
            if transforms is None:
                setattr(obj, attr_name, attr_value)
            else:
                setattr(obj, attr_name, transforms[1](attr_value))
        except AttributeError:
            # if its an ID we need to set the _id instead
            if attr_name == "id":
                attr_name = "_id"
                if transforms is None:
                    setattr(obj, attr_name, attr_value)
                else:
                    setattr(obj, attr_name, transforms[1](attr_value))

    return obj


def rs_get_protobuf_schema() -> GeneratedProtocolMessageType:
    return RecursiveSerde_PB



# stdlib
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from sympy.core.basic import Basic as BasicSymbol

# syft absolute
import syft as sy

# relative
from ....proto.core.adp.scalar_pb2 import (
    IntermediatePhiScalar as IntermediatePhiScalar_PB,
)
from ...common import UID
from ...common.serde.serializable import serializable
from ..entity import DataSubjectGroup
from ..entity import Entity
from .abstract.intermediate_scalar import IntermediateScalar
from .gamma_scalar import GammaScalar


@serializable()
class IntermediatePhiScalar(IntermediateScalar):
    """
    Serializable superclass for PhiScalars (Scalars with data from one entity).
    This is where all the functionality of a PhiScalar is implemented,
    such as searching for the max Lipshitz value.
    """

    def __init__(
        self,
        poly: BasicSymbol,
        entity: Union[Entity, DataSubjectGroup],
        id: Optional[UID] = None,
    ) -> None:
        super().__init__(poly=poly, id=id)
        self._gamma: Optional[GammaScalar] = None
        self.entity = entity

    def max_lipschitz_wrt_entity(
        self,
        entity: Entity,
    ) -> float:
        """Perform the search for max Lipshitz with respect to the current PhiScalar's entity"""
        return self.gamma.max_lipschitz_wrt_entity(entity=entity)

    @property
    def max_lipschitz(self) -> float:
        """Perform the search for max Lipschitz"""
        return self.gamma.max_lipschitz

    def __mul__(self, other: IntermediateScalar) -> IntermediateScalar:

        # relative
        from .intermediate_gamma_scalar import IntermediateGammaScalar

        if isinstance(
            other, IntermediateGammaScalar
        ):  # PhiScalar * GammaScalar = GammaScalar
            return self.gamma * other

        if not isinstance(
            other, IntermediatePhiScalar
        ):  # PhiScalar * Int/Float/etc = PhiScalar
            return IntermediatePhiScalar(poly=self.poly * other, entity=self.entity)

        # If the entities match, output is PhiScalar, otherwise it should be a GammaScalar
        if self.entity == other.entity:
            return IntermediatePhiScalar(
                poly=self.poly * other.poly, entity=self.entity
            )

        return (
            self.gamma * other.gamma
        )  # Phi(E1) * Phi(E2) = Gamma(E1, E2) | E1,E2 = Entities

    def __add__(self, other: IntermediateScalar) -> IntermediateScalar:

        # relative
        from .intermediate_gamma_scalar import IntermediateGammaScalar

        if isinstance(
            other, IntermediateGammaScalar
        ):  # PhiScalar + GammaScalar = GammaScalar
            return self.gamma + other

        if not isinstance(other, IntermediatePhiScalar):  # PhiScalar + 5 = PhiScalar
            return IntermediatePhiScalar(poly=self.poly + other, entity=self.entity)

        # if other is referencing the same individual
        if self.entity == other.entity:
            return IntermediatePhiScalar(
                poly=self.poly + other.poly, entity=self.entity
            )
        return (
            self.gamma + other.gamma
        )  # Phi(E1) * Phi(E2) = Gamma(E1, E2) | E1,E2 = Entities

    def __sub__(self, other: IntermediateScalar) -> IntermediateScalar:

        # relative
        from .intermediate_gamma_scalar import IntermediateGammaScalar

        if isinstance(other, IntermediateGammaScalar):
            return self.gamma - other

        # if other is a public value
        if not isinstance(other, IntermediatePhiScalar):
            return IntermediatePhiScalar(poly=self.poly - other, entity=self.entity)

        # if other is referencing the same individual
        if self.entity == other.entity:
            return IntermediatePhiScalar(
                poly=self.poly - other.poly, entity=self.entity
            )

        return (
            self.gamma - other.gamma
        )  # Phi(E1) * Phi(E2) = Gamma(E1, E2) | E1,E2 = Entities

    @property
    def gamma(self) -> GammaScalar:
        """Turn the PhiScalar into a GammaScalar, if another entity is involved"""
        if self._gamma is None:
            if (
                self.min_val is not None
                and self.value is not None
                and self.max_val is not None
            ):
                # TODO: Add prime to GammaScalar init
                self._gamma = GammaScalar(
                    min_val=self.min_val,
                    value=self.value,
                    max_val=self.max_val,
                    entity=self.entity,
                    prime=-1,  # TODO: shouldn't we be passing in some kind of prime here?
                )
            else:
                raise Exception("GammaScalar requires min_val, value and max_val")
        return self._gamma

    def _object2proto(self) -> IntermediatePhiScalar_PB:
        return IntermediatePhiScalar_PB(
            id=sy.serialize(self.id, to_proto=True),
            # gamma=sy.serialize(self._gamma, to_proto=True),
            # poly=self._poly if self._poly is not None else None,
            entity=sy.serialize(self.entity, to_proto=True),
        )

    @staticmethod
    def _proto2object(proto: IntermediatePhiScalar_PB) -> "IntermediatePhiScalar":
        intermediate_phi_scalar = IntermediatePhiScalar(
            id=sy.deserialize(proto.id, from_proto=True),
            entity=sy.deserialize(blob=proto.entity, from_proto=True),
            poly=None
            # poly=sy.deserialize(proto.poly)
        )
        # intermediate_phi_scalar._gamma = sy.deserialize(proto.gamma)
        return intermediate_phi_scalar

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return IntermediatePhiScalar_PB



# stdlib
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ...abstract.node_service_interface import NodeServiceInterface
from ..node_service.generic_payload.syft_message import NewSyftMessage as SyftMessage
from .permissions import BasePermission


class UserCanTriageRequest(BasePermission):
    # TODO: Exceptions could fail silently either depending upon the message
    # # or the permission (not sure which one, need to discuss)

    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        return (
            node.users.can_triage_requests(verify_key=verify_key)
            if verify_key
            else False
        )


class UserCanCreateUsers(BasePermission):
    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        return (
            node.users.can_create_users(verify_key=verify_key) if verify_key else False
        )


class UserCanEditRoles(BasePermission):
    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        return node.users.can_edit_roles(verify_key=verify_key) if verify_key else False


class UserCanUploadData(BasePermission):
    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        return (
            node.users.can_upload_data(verify_key=verify_key) if verify_key else False
        )


class IsNodeDaaEnabled(BasePermission):
    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        msg_payload = msg.payload  # type: ignore

        if node.setup.first(domain_name=node.name).daa and not msg_payload.daa_pdf:  # type: ignore
            return False

        return True


class NoRestriction(BasePermission):
    def has_permission(
        self,
        msg: SyftMessage,
        node: NodeServiceInterface,
        verify_key: Optional[VerifyKey],
    ) -> bool:
        return True


class UserIsOwner(BasePermission):
    def has_permission(
        self, msg: SyftMessage, node: NodeServiceInterface, verify_key: VerifyKey
    ) -> bool:

        if hasattr(msg.kwargs, "upcast"):
            msg_kwargs = msg.kwargs.upcast()  # type: ignore
        else:
            msg_kwargs = msg.kwargs

        user_id = msg_kwargs.get("user_id")

        if not user_id:
            return False

        _target_user = node.users.first(id=user_id)
        request_user = node.users.get_user(verify_key=verify_key)

        _is_owner = False
        if _target_user:  # If target user exists
            if (
                node.roles.first(id=_target_user.role).name
                == node.roles.owner_role.name
            ):  # If the user has role `Owner`
                _is_owner = True
            elif request_user and (
                _target_user.id == request_user.id
            ):  # request user is the target user
                _is_owner = True
        return _is_owner



# stdlib
from uuid import UUID

# third party
from numpy import ndarray

# relative
from ..tensor import AutogradTensor
from .op import Op


class AbsOp(Op):
    def forward(self, x: AutogradTensor) -> AutogradTensor:  # type: ignore
        self.x = x

        return AutogradTensor(x.child.__abs__(), requires_grad=x.requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:
            _grad = int(self.x > 0)  # returns 0s and 1s
            _grad = int((_grad * 2) - 1)  # returns -1s and 1s

            grad = _grad * grad

            self.x.add_grad(grad)

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# stdlib
from typing import Any
from typing import List as TypeList
from typing import Optional
from typing import Tuple as TypeTuple

# third party
import numpy as np
from scipy import optimize
from sympy.core.basic import Basic as BasicSymbol

# relative
from ...common import UID
from ..entity import Entity
from ..search import create_lookup_tables_for_symbol
from ..search import create_searchable_function_from_polynomial
from ..search import max_lipschitz_via_jacobian
from ..search import minimize_function
from ..search import ssid2obj
from .abstract.intermediate_scalar import IntermediateScalar
from .abstract.scalar import Scalar


class IntermediateGammaScalar(IntermediateScalar):
    """
    A Superclass for Scalars with data from multiple entities (GammaScalars).
    Most importantly, this is where all of the operations (+/-/*/div) are implemented,
    as well as the various methods with which to perform the search for the max Lipschitz.
    """

    def __init__(
        self,
        poly: BasicSymbol,
        min_val: float,
        max_val: float,
        id: Optional[UID] = None,
    ) -> None:
        super().__init__(poly=poly, id=id)
        self._min_val = float(min_val)
        self._max_val = float(max_val)
        self.is_linear: Optional[
            bool
        ] = None  # None means skip performance optimization

    # GammaScalar +/-/*/div other ---> GammaScalar
    def __add__(self, other: Any) -> IntermediateScalar:
        if isinstance(other, Scalar):
            # relative
            from .intermediate_phi_scalar import IntermediatePhiScalar

            if isinstance(other, IntermediatePhiScalar):
                other = other.gamma
            return IntermediateGammaScalar(
                poly=self.poly + other.poly,
                min_val=self.min_val + other.min_val,
                max_val=self.max_val + other.max_val,
            )
        return IntermediateGammaScalar(
            poly=self.poly + other,
            min_val=self.min_val + other,
            max_val=self.max_val + other,
        )

    def __sub__(self, other: Any) -> IntermediateScalar:
        if isinstance(other, Scalar):
            # relative
            from .intermediate_phi_scalar import IntermediatePhiScalar

            if isinstance(other, IntermediatePhiScalar):
                other = other.gamma
            return IntermediateGammaScalar(
                poly=self.poly - other.poly,
                min_val=self.min_val - other.min_val,
                max_val=self.max_val - other.max_val,
            )
        return IntermediateGammaScalar(
            poly=self.poly - other,
            min_val=self.min_val - other,
            max_val=self.max_val - other,
        )

    def __mul__(self, other: Any) -> IntermediateScalar:
        if isinstance(other, Scalar):
            # relative
            from .intermediate_phi_scalar import IntermediatePhiScalar

            if isinstance(other, IntermediatePhiScalar):
                other = other.gamma

            max_val = max(
                self.min_val * other.min_val,
                self.min_val * other.max_val,
                self.max_val * other.min_val,
                self.max_val * other.max_val,
            )

            min_val = min(
                self.min_val * other.min_val,
                self.min_val * other.max_val,
                self.max_val * other.min_val,
            )

            return IntermediateGammaScalar(
                poly=self.poly * other.poly, max_val=max_val, min_val=min_val
            )

        max_val = max(
            self.min_val * other,
            self.max_val * other,
        )

        min_val = min(
            self.min_val * other,
            self.max_val * other,
        )

        return IntermediateGammaScalar(
            poly=self.poly * other, min_val=min_val, max_val=max_val
        )

    def max_lipschitz_via_explicit_search(
        self, force_all_searches: bool = False
    ) -> TypeTuple[TypeList[optimize.OptimizeResult], np.float64]:

        r1 = np.array([x.poly for x in self.input_scalars])  # type: ignore

        # relative
        from .gamma_scalar import GammaScalar

        r2_diffs = np.array(
            [
                GammaScalar(x.min_val, x.value, x.max_val, entity=x.entity, prime=x.prime).poly  # type: ignore
                for x in self.input_scalars
            ]
        )
        r2 = r1 + r2_diffs

        fr1 = self.poly
        fr2 = self.poly.copy().subs({x[0]: x[1] for x in list(zip(r1, r2))})

        left = np.sum(np.square(fr1 - fr2)) ** 0.5
        right = np.sum(np.square(r1 - r2)) ** 0.5

        C = -left / right

        i2s, s2i = create_lookup_tables_for_symbol(C)
        search_fun = create_searchable_function_from_polynomial(
            poly=C, symbol2index=s2i
        )

        r1r2diff_zip = list(zip(r1, r2_diffs))

        s2range = {}
        for _input_scalar, _additive_counterpart in r1r2diff_zip:
            input_scalar = ssid2obj[_input_scalar.name]
            additive_counterpart = ssid2obj[_additive_counterpart.name]

            s2range[input_scalar.ssid] = (input_scalar.min_val, input_scalar.max_val)
            s2range[additive_counterpart.ssid] = (
                input_scalar.min_val,
                input_scalar.max_val,
            )

        rranges = list()
        for _, symbol in enumerate(i2s):
            rranges.append(s2range[symbol])

        r2_indices_list = list()
        min_max_list = list()
        for r2_val in r2:
            r2_syms = [ssid2obj[x.name] for x in r2_val.free_symbols]
            r2_indices = [s2i[x.ssid] for x in r2_syms]

            r2_indices_list.append(r2_indices)
            min_max_list.append((r2_syms[0].min_val, r2_syms[0].max_val))

        functions = list()
        for i in range(2):
            f1 = (
                lambda x, i=i: x[r2_indices_list[i][0]]
                + x[r2_indices_list[i][1]]
                + min_max_list[i][0]
            )
            f2 = (
                lambda x, i=i: -(x[r2_indices_list[i][0]] + x[r2_indices_list[i][1]])
                + min_max_list[i][1]
            )

            functions.append(f1)
            functions.append(f2)

        constraints = [{"type": "ineq", "fun": f} for f in functions]

        def non_negative_additive_terms(symbol_vector: np.ndarray) -> np.float64:
            out = 0
            for index in [s2i[x.name] for x in r2_diffs]:
                out += symbol_vector[index] ** 2
            # there's a small bit of rounding error from this constraint - this should
            # only be used as a double check or as a backup!!!
            return out**0.5 - 1 / 2**16

        constraints.append({"type": "ineq", "fun": non_negative_additive_terms})
        results = minimize_function(
            f=search_fun,
            rranges=rranges,
            constraints=constraints,
            force_all_searches=force_all_searches,
        )

        return results, C

    def max_lipschitz_via_jacobian(
        self,
        input_entity: Optional[Entity] = None,
        data_dependent: bool = True,
        force_all_searches: bool = False,
        try_hessian_shortcut: bool = False,
    ) -> TypeList[optimize.OptimizeResult]:
        return max_lipschitz_via_jacobian(
            scalars=[self],
            input_entity=input_entity,
            data_dependent=data_dependent,
            force_all_searches=force_all_searches,
            try_hessian_shortcut=try_hessian_shortcut,
        )  # type: ignore

    @property
    def max_lipschitz(self) -> float:
        result = self.max_lipschitz_via_jacobian()[0][-1]
        if isinstance(result, float):
            return -result
        else:
            return -float(result.fun)

    def max_lipschitz_wrt_entity(self, entity: Entity) -> float:
        result = self.max_lipschitz_via_jacobian(input_entity=entity)[0][-1]
        if isinstance(result, float):
            return -result
        else:
            return -float(result.fun)



# stdlib
from uuid import UUID

# third party
from numpy import ndarray

# relative
from .....core.common.serde.serializable import serializable
from ...passthrough import is_acceptable_simple_type  # type: ignore
from ..tensor import AutogradTensor
from .op import Op


@serializable(recursive_serde=True)
class MulOp(Op):
    """Multiplication operation with 2 tensors"""

    __attr_allowlist__ = [
        "x",
        "y",
    ]

    def forward(self, x: AutogradTensor, y: AutogradTensor) -> AutogradTensor:  # type: ignore
        self.x = x
        self.y = y

        requires_grad = x.requires_grad

        if is_acceptable_simple_type(y):
            return AutogradTensor(x.child * y, requires_grad=requires_grad)

        # print(y)
        requires_grad = requires_grad or y.requires_grad

        # print()
        # print()
        # print("mul._backward")
        # print(x.child)
        # print()
        # print(y.child)
        if is_acceptable_simple_type(y.child):
            return AutogradTensor(x.child * y.child, requires_grad=requires_grad)

        return AutogradTensor(y.child * x.child, requires_grad=requires_grad)

    def _backward(self, grad: ndarray, backprop_id: UUID) -> None:

        y_is_simple = is_acceptable_simple_type(self.y)

        if self.x.requires_grad:

            if y_is_simple:
                self.x.add_grad(grad * self.y)
            else:
                temp = self.y * grad
                self.x.add_grad(temp)

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)

        # if y_is_simple then it's definitely not an autograd tensor (doesn't need to be
        # backpropagated into. also if it doesn't .requires_grad
        if not y_is_simple and self.y.requires_grad:
            self.y.add_grad(self.x * grad)
            if self.y.grad_fn:
                self.y.backward(backprop_id=backprop_id)



# stdlib
from typing import Any

# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String

# relative
from . import Base
from ..... import deserialize
from ..... import serialize
from ....adp.idp_gaussian_mechanism import iDPGaussianMechanism


class Mechanism(Base):
    __tablename__ = "mechanism"

    id = Column(Integer(), primary_key=True, autoincrement=True)

    # "Kritika the data scientist"
    user_key = Column(LargeBinary(2048), default="")

    # the name of the medical patient whose private information was queried
    entity_name = Column(String(1024), default="")

    # list of mechanisms about medical patients = [Bob, Sue, James]
    mechanism_bin = Column(LargeBinary(3072), default=None)

    @property
    def obj(self) -> Any:
        _obj = deserialize(self.mechanism_bin, from_bytes=True)  # TODO: techdebt fix
        # iDPGaussianMechanism.__new__ used by the recursive serde
        # will not initialize iDPGaussianMechanism super class.
        # Since we're extending a third party lib class to perform dp (autodp lib)
        # we need to force super class initialization in order to provide support for
        # autodp internal methods created in execution time (e.g. RenyiDP)
        # if len(obj_list):
        iDPGaussianMechanism.__init__(
            self=_obj,
            sigma=_obj.params["sigma"],
            squared_l2_norm=_obj.params["private_value"],
            squared_l2_norm_upper_bound=_obj.params["public_value"],
            L=_obj.params["L"],
            entity_name=_obj.entity_name,
            user_key=_obj.user_key,
        )

        return _obj

    @obj.setter
    def obj(self, value: Any) -> None:
        self.mechanism_bin = serialize(value, to_bytes=True)  # TODO: techdebt fix



# future
from __future__ import annotations

# stdlib
from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from primesieve.numpy import primes

# relative
from ...logger import warning
from ...proto.core.adp.scalar_manager_pb2 import (
    VirtualMachinePrivateScalarManager as VirtualMachinePrivateScalarManager_PB,
)
from ...proto.core.adp.scalar_manager_pb2 import PrimeFactory as PrimeFactory_PB
from ..common.serde.deserialize import _deserialize as deserialize
from ..common.serde.serializable import serializable
from ..common.serde.serialize import _serialize as serialize
from .entity import Entity
from .scalar.gamma_scalar import GammaScalar

prime_exp_cache = {}


def get_cached_primes(total: int) -> list:
    global prime_exp_cache
    if total not in prime_exp_cache:
        prime_exp_cache[total] = primes(total)
    return prime_exp_cache[total]


@serializable()
class PrimeFactory:
    """IMPORTANT: it's very important that two tensors be able to tell that
    they are indeed referencing the EXACT same PrimeFactory. At present this is done
    by ensuring that it is literally the same python object. In the future, we will
    probaby need to formalize this. However, the main way this could go wrong is if we
    created some alternate way for checking to see if two prime factories 'sortof looked
    the same' but which in fact weren't the EXACT same object. This could lead to
    security leaks wherein two tensors think two different symbols in fact are the
    same symbol."""

    def __init__(self) -> None:
        self.exp = 2
        self.prime_numbers: list = get_cached_primes(total=10**self.exp)

    def get(self, index: int) -> int:
        while index > len(self.prime_numbers) - 1:
            self.exp += 1
            self.prime_numbers = get_cached_primes(total=10**self.exp)
        return self.prime_numbers[index]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PrimeFactory):
            return (
                self.exp == other.exp
            )  # Since the list is sorted, we can just compare the last value
        return self == other

    def _object2proto(self) -> PrimeFactory_PB:
        return PrimeFactory_PB()

    @staticmethod
    def _proto2object(
        proto: PrimeFactory_PB,
    ) -> PrimeFactory:
        return PrimeFactory()

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return PrimeFactory_PB


@serializable()
class VirtualMachinePrivateScalarManager:
    __slots__ = ("prime_factory", "prime2symbol", "hash_cache", "index")

    def __init__(
        self,
        prime_factory: Optional[PrimeFactory] = None,
        prime2symbol: Optional[Dict[Any, Any]] = None,
    ) -> None:
        self.prime_factory = (
            prime_factory if prime_factory is not None else PrimeFactory()
        )

        if prime2symbol is None:
            prime2symbol = {}

        self.prime2symbol = prime2symbol
        self.hash_cache: Optional[int] = None
        self.index = 0

    def get_symbol(
        self,
        min_val: Union[bool, int, float],
        value: Union[bool, int, float],
        max_val: Union[bool, int, float],
        entity: Entity,
    ) -> int:
        # NOTE: this is overly conservative because it always creates a new scalar even when
        # a computationally equivalent one might exist somewhere already.

        gs = GammaScalar(
            min_val=min_val,
            value=value,
            max_val=max_val,
            entity=entity,
            prime=int(self.prime_factory.get(self.index)),
        )

        self.prime2symbol[int(gs.prime)] = gs
        self.hash_cache = None
        self.index += 1
        return gs.prime

    def __hash__(self) -> int:
        # TODO: safely invalidate the hash_cache when prime2symbol is mutated
        if self.hash_cache is None:
            prime2symbol = frozenset(self.prime2symbol.items())
            self.hash_cache = hash(self.prime_factory.exp) + hash(prime2symbol)
        return self.hash_cache

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, VirtualMachinePrivateScalarManager):
            return (
                self.prime_factory == other.prime_factory
                and self.prime2symbol == other.prime2symbol
            )
        return self == other

    @property
    def primes_allocated(self) -> list:
        return list(self.prime2symbol.keys())

    def copy(self) -> VirtualMachinePrivateScalarManager:

        # intentionally avoiding accidentally copying the prime factory
        new_mgr = VirtualMachinePrivateScalarManager(
            prime2symbol=deepcopy(self.prime2symbol)
        )

        return new_mgr

    def _object2proto(self) -> VirtualMachinePrivateScalarManager_PB:
        return VirtualMachinePrivateScalarManager_PB(
            prime_factory=serialize(self.prime_factory),
            prime2symbol=serialize(self.prime2symbol),
        )

    @staticmethod
    def _proto2object(
        proto: VirtualMachinePrivateScalarManager_PB,
    ) -> VirtualMachinePrivateScalarManager:
        return VirtualMachinePrivateScalarManager(
            prime_factory=deserialize(proto.prime_factory),
            prime2symbol=deserialize(proto.prime2symbol),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return VirtualMachinePrivateScalarManager_PB

    def combine_(self, vsm2: VirtualMachinePrivateScalarManager) -> None:
        """
        Combine two ScalarManagers by merging their prime factories.
        ASSUME: vsm1 is the source of truth; we won't be changing its prime numbers
        """

        # To check: vsm2.prime2symbol size doesn't change
        if id(self.prime2symbol) != id(vsm2.prime2symbol):
            copy_dict = vsm2.prime2symbol.copy()
            length = len(self.primes_allocated)
            for prime_number, gs in copy_dict.items():
                if int(prime_number) in self.prime2symbol:  # If there's a collision
                    new_prime = self.prime_factory.get(length)
                    gs.prime = new_prime
                    self.prime2symbol[int(new_prime)] = gs
                    del vsm2.prime2symbol[int(prime_number)]
                    vsm2.prime2symbol[int(new_prime)] = gs
                    length += 1
                else:
                    self.prime2symbol[int(gs.prime)] = gs
        else:
            warning("Detected prime2symbol where two tensors were using the same dict")



"""Route is an abstraction between Connection and Client.
A Client represents an API for interfacing with a Node of
a particular type at a specific location. We assume that
a Client may have multiple ways of interacting with a node
(such as multiple protocols, http, websockets, broadcast server
, etc.). We assume that a Client also has knowledge of the
full variety of messages and message types which can be sent
to a Node. It is the highest level of abstraction.

A connection is the lowest level of abstraction, representing
a real network interface to a single destination. Importantly,
this destination may not even be the ultimate destination
for a message. In this sense, a connection is just a single "hop"
in a potentially long series of hops between a Client and
a Node.

However, it's important to note that a Client doesn't need to
have direct access to the node it communicates with because
Nodes have a variety of services for forwarding messages
around. For example, if I have two clients, one pointing
to a Domain node and another pointing to a Device within
the Domain, both might use the same connection object
because all messages to a device have to go through a domain.
This is because, for permissions reasons, the domain represents
an intentional information bottleneck.

This might beg the question, what is the point of Route if
the right combination of Clients and Connections can always
get a message to where it needs to go?

The answer is that neither a Connection, Client, or Node are
appropriate for the nature of four combined constraints.

1) the existence of multiple connections from client -> node
with varying tradeoffs (websocket, http, webRTC, gRPC, etc.)
2) the existence of multiple clients which share the same
connection because of intentional information bottlenecks.
3) the existence of 1->many connection types (pub-sub)
4) the varying cost of different networks (edge cellular, and
cloud networking costs being the biggest)

We need the ability for an arbitrary client (which represents
an abstract, indirect connection to a specific remote node of
a specific type) to be able to consider a message, its address
, and other metadata, and make a logical decision on which connection
should be used. (For example, some messages block for a reply)

We need the connection decision to take into account the latency,
connection health, and overall cost of each connection option as
measured by fixed constants as well as previous uses of the
connection.

This necessitates an abstraction which lets us store connection
information both about hops we know about and about hops we don't
know about.

This abstraction is called a "Route" and each hop is called a "Hop"
where any collection of hops within the route is called a
"RouteSegment". Note that a Route need only be initialized with
its source and destination(s).

When a route is used by a client, it is used to decide which route
would be best to take for a particular message.

When a route is attached to a message, it implies the expected
journey the message is to take to the address. A node may or may
not actually respect this journey.

Routes can be created locally, but they are best created by asking
the network for a good route to a remote worker which allows for
things like network speed to be interrogated in the creation of
a route.

While in theory, this data-structure could allow for very sophisticated
network analysis, in the beginning, it will mostly exist to choose
between Pub-sub, Request-Response, and Streaming options. In the
long run it will allow us to design routes which explicitly
avoid the most costly network bottlenecks. For example, if Grid is
advertising a new Federated Learning job, instead of sending the
model directly to 10,000 clients (lots of bandwidth costs on the
cloud node), it could send it to one cellphone with WiFi which
would have instructions for how to forward to a binary tree of other
node, propagating the model to all node which asked for it.
"""

# stdlib
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# relative
from ...logger import debug
from ...logger import traceback_and_raise
from ...proto.core.io.route_pb2 import SoloRoute as SoloRoute_PB
from ..common.message import SignedEventualSyftMessageWithoutReply
from ..common.message import SignedImmediateSyftMessageWithReply
from ..common.message import SignedImmediateSyftMessageWithoutReply
from ..common.object import ObjectWithID
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import serializable
from ..common.serde.serialize import _serialize
from .connection import BidirectionalConnection
from .connection import ClientConnection
from .location import Location
from .location import SpecificLocation
from .virtual import VirtualClientConnection


class RouteSchema(ObjectWithID):
    """An object which contains the IDs of the origin node and
    set of the destination node. Multiple routes can subscribe
    to the same RouteSchema and routing, logic is thus split into
    two groups of functionality:

    1) Discovering new routes
    2) Comparing known routes to find the best one for a message
    """

    def __init__(self, destination: Location):
        self.destination = destination


class Route(ObjectWithID):
    def __init__(self, schema: RouteSchema, stops: Optional[list] = None):
        super().__init__()
        if stops is None:
            stops = list()
        self.schema = schema
        self.stops = stops

    @property
    def icon(self) -> str:
        return "🛣️ "

    @property
    def pprint(self) -> str:
        return f"{self.icon} ({str(self.__class__.__name__)})"

    def send_immediate_msg_without_reply(
        self,
        msg: SignedImmediateSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)

    def send_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply, timeout: Optional[float] = None
    ) -> SignedImmediateSyftMessageWithoutReply:
        traceback_and_raise(NotImplementedError)

    def send_eventual_msg_without_reply(
        self,
        msg: SignedEventualSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        traceback_and_raise(NotImplementedError)


@serializable()
class SoloRoute(Route):
    def __init__(
        self,
        destination: Location,
        connection: Union[ClientConnection, BidirectionalConnection],
    ) -> None:
        super().__init__(schema=RouteSchema(destination=destination))
        self.connection = connection

    def send_immediate_msg_without_reply(
        self,
        msg: SignedImmediateSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        debug(f"> Routing {msg.pprint} via {self.pprint}")
        self.connection.send_immediate_msg_without_reply(msg=msg, timeout=timeout)

    def send_eventual_msg_without_reply(
        self,
        msg: SignedEventualSyftMessageWithoutReply,
        timeout: Optional[float] = None,
    ) -> None:
        self.connection.send_eventual_msg_without_reply(msg=msg, timeout=timeout)

    def send_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply, timeout: Optional[float] = None
    ) -> SignedImmediateSyftMessageWithoutReply:
        return self.connection.send_immediate_msg_with_reply(msg=msg, timeout=timeout)

    def _object2proto(self) -> SoloRoute_PB:
        route = SoloRoute_PB(
            id=self.id._object2proto(),
            destination=self.schema.destination._object2proto(),
        )
        if isinstance(self.connection, VirtualClientConnection):
            route.virtual_connection.CopyFrom(
                _serialize(self.connection, to_proto=True)
            )
        else:
            route.grid_connection.CopyFrom(_serialize(self.connection, to_proto=True))
        return route

    @staticmethod
    def _proto2object(proto: SoloRoute_PB) -> "SoloRoute":
        if proto.HasField("virtual_connection"):
            connection = VirtualClientConnection._proto2object(proto.virtual_connection)
        elif proto.HasField("grid_connection"):
            connection = _deserialize(proto.grid_connection)
        else:
            raise TypeError("Unsupported type of connection")

        route = SoloRoute(
            destination=SpecificLocation._proto2object(proto.destination),
            connection=connection,
        )
        route._id = _deserialize(proto.id)
        return route

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SoloRoute_PB


class BroadcastRoute(SoloRoute):
    """
    A route used for pub/sub type systems.
    """

    def __init__(self, destination: Location, connection: ClientConnection) -> None:
        super().__init__(destination=destination, connection=connection)
        # self.connection.topic = destination.topic



# stdlib
from typing import Tuple
from typing import Union

# third party
import numpy as np
import pyarrow as pa
import torch

# relative
from ...core.common.serde.serializable import serializable
from ...experimental_flags import ApacheArrowCompression
from ...experimental_flags import flags
from ...proto.lib.numpy.array_pb2 import NumpyProto
from ..torch.tensor_util import tensor_deserializer
from ..torch.tensor_util import tensor_serializer

SUPPORTED_BOOL_TYPES = [np.bool_]
SUPPORTED_INT_TYPES = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

SUPPORTED_FLOAT_TYPES = [
    np.float16,
    np.float32,
    np.float64,
]

SUPPORTED_DTYPES = SUPPORTED_BOOL_TYPES + SUPPORTED_INT_TYPES + SUPPORTED_FLOAT_TYPES

DTYPE_REFACTOR = {
    np.dtype("uint16"): np.int16,
    np.dtype("uint32"): np.int32,
    np.dtype("uint64"): np.int64,
}


def arrow_serialize(
    obj: np.ndarray, get_bytes: bool = False
) -> Union[Tuple[bytes, int], NumpyProto]:
    original_dtype = obj.dtype
    apache_arrow = pa.Tensor.from_numpy(obj=obj)
    sink = pa.BufferOutputStream()
    pa.ipc.write_tensor(apache_arrow, sink)
    buffer = sink.getvalue()
    if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
        numpy_bytes = buffer.to_pybytes()
    else:
        numpy_bytes = pa.compress(
            buffer, asbytes=True, codec=flags.APACHE_ARROW_COMPRESSION.value
        )
    dtype = original_dtype.name

    if get_bytes:
        return numpy_bytes, buffer.size
    else:
        return NumpyProto(
            arrow_data=numpy_bytes, dtype=dtype, decompressed_size=buffer.size
        )


def arrow_deserialize(
    buf: bytes, decompressed_size: int, dtype: str = "int32"
) -> np.ndarray:
    str_dtype = dtype
    original_dtype = np.dtype(str_dtype)
    if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
        reader = pa.BufferReader(buf)
        buf = reader.read_buffer()
    else:
        buf = pa.decompress(
            buf,
            decompressed_size=decompressed_size,
            codec=flags.APACHE_ARROW_COMPRESSION.value,
        )

    result = pa.ipc.read_tensor(buf)
    np_array = result.to_numpy()
    np_array.setflags(write=True)
    return np_array.astype(original_dtype)


def protobuf_serialize(obj: np.ndarray) -> NumpyProto:
    original_dtype = obj.dtype
    if original_dtype not in SUPPORTED_DTYPES:
        raise NotImplementedError(f"{original_dtype} is not supported")

    if original_dtype in DTYPE_REFACTOR:
        # store as a signed int, the negative wrap around values convert back to the
        # same original unsigned values on the other side
        obj = obj.astype(DTYPE_REFACTOR[original_dtype])

    # Cloning seems to cause the worker to freeze if the array is larger than around
    # 800k in data and since we are serializing it immediately afterwards I don't
    # think its needed anyway
    # tensor = torch.from_numpy(obj).clone()
    tensor = torch.from_numpy(obj)
    tensor_bytes = tensor_serializer(tensor)
    dtype = original_dtype.name
    return NumpyProto(proto_data=tensor_bytes, dtype=dtype)


def protobuf_deserialize(proto: NumpyProto) -> np.ndarray:
    tensor = tensor_deserializer(proto.proto_data)
    array = tensor.to("cpu").detach().numpy().copy()
    str_dtype = proto.dtype
    original_dtype = np.dtype(str_dtype)
    obj = array.astype(original_dtype)
    return obj


def serialize_numpy_array(obj: np.ndarray) -> NumpyProto:
    if flags.APACHE_ARROW_TENSOR_SERDE:
        return arrow_serialize(obj)
    else:
        return protobuf_serialize(obj)


def deserialize_numpy_array(proto: NumpyProto) -> np.ndarray:
    if proto.HasField("arrow_data"):
        return arrow_deserialize(proto.arrow_data, proto.decompressed_size)
    else:
        return protobuf_deserialize(proto)


serializable(generate_wrapper=True)(
    wrapped_type=np.ndarray,
    import_path="numpy.ndarray",
    protobuf_scheme=NumpyProto,
    type_object2proto=serialize_numpy_array,
    type_proto2object=deserialize_numpy_array,
)



"""This module contains `Module` attribute representing a module which contains other models or callables."""

# stdlib
import inspect
import sys
from types import ModuleType
from typing import Any
from typing import Callable as CallableT
from typing import List
from typing import Optional
from typing import Union

# relative
from .. import ast
from ..logger import traceback_and_raise
from .callable import Callable


def is_static_method(host_object, attr) -> bool:  # type: ignore
    """Check if attribute `attr` of an object is a static method.

    Args:
        host_object: The object.
        attr: The attribute of the object.

    Returns:
        bool: `True` if `host_object[attr]` is a static method.
    """
    value = getattr(host_object, attr)

    # Host object must contain the method resolution order attribute (mro)
    if not hasattr(host_object, "__mro__"):
        return False

    for cls in inspect.getmro(host_object):
        if inspect.isroutine(value):
            if attr in cls.__dict__:
                bound_value = cls.__dict__[attr]
                if isinstance(bound_value, staticmethod):
                    return True
    return False


class Module(ast.attribute.Attribute):
    """A module which contains other modules or callables."""

    def __init__(
        self,
        client: Optional[Any],
        parent: Optional[ast.attribute.Attribute] = None,
        path_and_name: Optional[str] = None,
        object_ref: Optional[Union[CallableT, ModuleType]] = None,
        return_type_name: Optional[str] = None,
    ):
        """Base constructor for Module Attribute.

        Args:
            client: The client for which all computation is being executed.
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            parent: The parent node of the module.
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The given action's return type name, with its full path, in string format.
        """
        super().__init__(
            path_and_name=path_and_name,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )

        if object_ref is None and self.name:
            try:
                self.object_ref = sys.modules[path_and_name if path_and_name else ""]
            except Exception:
                self.object_ref = getattr(self.parent.object_ref, self.name)

    def add_attr(
        self,
        attr_name: str,
        attr: Optional[Union[Callable, CallableT]],
        is_static: bool = False,
    ) -> None:
        """Add an attribute to the current module.

        Args:
            attr_name: The name of the attribute, e.g. `List` of the path `syft.lib.python.List`.
            attr: The attribute object.
            is_static: The actual Python object for which the computation is being made.
        """
        self.__setattr__(attr_name, attr)

        if is_static is True:
            traceback_and_raise(
                ValueError("Static methods shouldn't be added to an object.")
            )

        if attr is None:
            traceback_and_raise(ValueError("An attribute reference has to be passed."))

        # If `add_attr` is called directly, we need to cache the path as well
        attr_ref = getattr(attr, "object_ref", None)
        path = getattr(attr, "path_and_name", None)
        if attr_ref not in self.lookup_cache and path is not None:
            self.lookup_cache[attr_ref] = path

        self.attrs[attr_name] = attr  # type: ignore

    def __call__(
        self,
        path: Union[List[str], str],
        index: int = 0,
        obj_type: Optional[type] = None,
    ) -> Optional[Union[Callable, CallableT]]:
        """Execute the given node object reference.

        Args:
            path: The node path in module to execute, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List"].
            index : The associated position in the path for the current node.
            obj_type: The type of the object to be called, whose path is resolved from `lookup_cache`.

        Returns:
            The results of running the computation on the object reference.
        """
        self.apply_node_changes()

        if obj_type is not None:
            if obj_type in self.lookup_cache:
                path = self.lookup_cache[obj_type]

        _path: List[str] = (
            path.split(".") if isinstance(path, str) else path if path else []
        )

        if not _path:
            traceback_and_raise(
                ValueError("Can't execute remote call if path is not specified.")
            )

        resolved = self.attrs[_path[index]](
            path=_path, index=index + 1, obj_type=obj_type
        )

        return resolved

    def __repr__(self) -> str:
        """Tree view of the `module`.

        Returns:
            The string representation of the `Module` Attribute.
        """
        out = "Module:\n"
        for name, module in self.attrs.items():
            out += "\t." + name + " -> " + str(module).replace("\t.", "\t\t.") + "\n"

        return out

    def add_path(
        self,
        path: List[str],
        index: int = 0,
        return_type_name: Optional[str] = None,
        framework_reference: Optional[ModuleType] = None,
        is_static: bool = False,
    ) -> None:
        """The add_path method adds new nodes in AST based on type of current node and type of object to be added.

        Args:
            path: The node path added in AST, e.g. `syft.lib.python.List` or ["syft", "lib", "python", "List].
            index: The associated position in the path for the current node.
            framework_reference: The Python framework in which we can resolve same path to obtain Python object.
            return_type_name: The return type name of the given action as a string with its full path.
            is_static: If the queried object is static, it has to be found on AST itself, not on an existing pointer.
        """
        if index >= len(path):
            return

        if path[index] not in self.attrs:
            attr_ref = getattr(self.object_ref, path[index])

            if inspect.ismodule(attr_ref):
                self.add_attr(
                    attr_name=path[index],
                    attr=ast.module.Module(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        parent=self,
                    ),
                )
            elif inspect.isclass(attr_ref):
                klass = ast.klass.Class(
                    path_and_name=".".join(path[: index + 1]),
                    object_ref=attr_ref,
                    return_type_name=return_type_name,
                    client=self.client,
                    parent=self,
                )
                self.add_attr(
                    attr_name=path[index],
                    attr=klass,
                )
            elif inspect.isfunction(attr_ref) or inspect.isbuiltin(attr_ref):
                is_static = is_static_method(self.object_ref, path[index])

                self.add_attr(
                    attr_name=path[index],
                    attr=ast.callable.Callable(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        is_static=is_static,
                        parent=self,
                    ),
                )
            elif inspect.isdatadescriptor(attr_ref):
                self.add_attr(
                    attr_name=path[index],
                    attr=ast.property.Property(
                        path_and_name=".".join(path[: index + 1]),
                        object_ref=attr_ref,
                        return_type_name=return_type_name,
                        client=self.client,
                        parent=self,
                    ),
                )
            elif index == len(path) - 1:
                static_attribute = ast.static_attr.StaticAttribute(
                    path_and_name=".".join(path[: index + 1]),
                    return_type_name=return_type_name,
                    client=self.client,
                    parent=self,
                )
                setattr(self, path[index], static_attribute)
                self.attrs[path[index]] = static_attribute
                return

        attr = self.attrs[path[index]]
        attr_ref = getattr(self.object_ref, path[index], None)
        if attr_ref is not None and attr_ref not in self.lookup_cache:
            self.lookup_cache[attr_ref] = path

        attr.add_path(path=path, index=index + 1, return_type_name=return_type_name)

    def __getattribute__(self, item: str) -> Any:
        """Get Attribute of a `module`.

        Args:
            item: Attribute.

        Returns:
            The value of the attribute.
        """
        target_object = super().__getattribute__(item)
        if isinstance(target_object, ast.static_attr.StaticAttribute):
            return target_object.get_remote_value()
        return target_object

    def __setattr__(self, key: str, value: Any) -> None:
        """Set atttribute of a module.

        Args:
            key: name of attribute to change.
            value: value to change attribute `key` to.
        """
        if hasattr(super(), "attrs"):
            attrs = super().__getattribute__("attrs")
            if key in attrs:
                target_object = self.attrs[key]
                if isinstance(target_object, ast.static_attr.StaticAttribute):
                    return target_object.set_remote_value(value)
        return super().__setattr__(key, value)

    def fetch_live_object(self) -> object:
        """Get the new object and its attributes from the client.

        Returns:
            The updated object from the client.
        """
        try:
            return sys.modules[self.path_and_name if self.path_and_name else ""]
        except Exception:
            return getattr(self.parent.object_ref, self.name)



# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy.orm import Session

# relative
from . import Base


class StorageMetadata(Base):
    __tablename__ = "storage_metadata"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    length = Column(Integer())

    def __str__(self) -> str:
        return f"<StorageMetadata length: {self.length}>"


def get_metadata(db_session: Session) -> StorageMetadata:
    metadata = db_session.query(StorageMetadata).first()

    if metadata is None:
        metadata = StorageMetadata(length=0)
        db_session.add(metadata)
        db_session.commit()

    return metadata



# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


# this table holds the list of known nodes usually peer domains
class Node(Base):
    __tablename__ = "node"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    node_uid = Column(String(255))
    node_name = Column(String(255))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/single_entity_phi_tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.adp import entity_pb2 as proto_dot_core_dot_adp_dot_entity__pb2
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.lib.numpy import array_pb2 as proto_dot_lib_dot_numpy_dot_array__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n0proto/core/tensor/single_entity_phi_tensor.proto\x12\x10syft.core.tensor\x1a\x1bproto/core/adp/entity.proto\x1a\x1bproto/lib/numpy/array.proto\x1a\x1bproto/core/io/address.proto\x1a%proto/core/common/common_object.proto\"\x82\x03\n)TensorWrappedSingleEntityPhiTensorPointer\x12%\n\x06\x65ntity\x18\x01 \x01(\x0b\x32\x15.syft.core.adp.Entity\x12,\n\x08min_vals\x18\x02 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12,\n\x08max_vals\x18\x03 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProto\x12'\n\x08location\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x16\n\x0escalar_manager\x18\x05 \x01(\x0c\x12-\n\x0eid_at_location\x18\x06 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bobject_type\x18\x07 \x01(\t\x12\x0c\n\x04tags\x18\x08 \x03(\t\x12\x13\n\x0b\x64\x65scription\x18\t \x01(\t\x12\x14\n\x0cpublic_shape\x18\n \x01(\x0c\x12\x14\n\x0cpublic_dtype\x18\x0b \x01(\x0c\x62\x06proto3"
)


_TENSORWRAPPEDSINGLEENTITYPHITENSORPOINTER = DESCRIPTOR.message_types_by_name[
    "TensorWrappedSingleEntityPhiTensorPointer"
]
TensorWrappedSingleEntityPhiTensorPointer = _reflection.GeneratedProtocolMessageType(
    "TensorWrappedSingleEntityPhiTensorPointer",
    (_message.Message,),
    {
        "DESCRIPTOR": _TENSORWRAPPEDSINGLEENTITYPHITENSORPOINTER,
        "__module__": "proto.core.tensor.single_entity_phi_tensor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.tensor.TensorWrappedSingleEntityPhiTensorPointer)
    },
)
_sym_db.RegisterMessage(TensorWrappedSingleEntityPhiTensorPointer)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TENSORWRAPPEDSINGLEENTITYPHITENSORPOINTER._serialized_start = 197
    _TENSORWRAPPEDSINGLEENTITYPHITENSORPOINTER._serialized_end = 583
# @@protoc_insertion_point(module_scope)



# stdlib
from collections import OrderedDict as PyOrderedDict
from collections.abc import ItemsView
from collections.abc import KeysView
from collections.abc import ValuesView
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ....core.common.serde.serializable import serializable
from ....core.common.uid import UID
from ....logger import traceback_and_raise
from ....proto.lib.python.collections.ordered_dict_pb2 import (
    OrderedDict as OrderedDict_PB,
)
from ..iterator import Iterator
from ..primitive_factory import PrimitiveFactory
from ..primitive_factory import isprimitive
from ..primitive_interface import PyPrimitive
from ..types import SyPrimitiveRet
from ..util import downcast
from ..util import upcast


@serializable()
class OrderedDict(PyOrderedDict, PyPrimitive):
    def __init__(self, *args: Any, _id: UID = UID(), **kwds: Any):
        super().__init__(*args, **kwds)
        self._id = _id

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def __contains__(self, other: Any) -> SyPrimitiveRet:
        res = super().__contains__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __delitem__(self, other: Any) -> None:
        res = super().__delitem__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = super().__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __getitem__(self, other: Any) -> SyPrimitiveRet:
        res = super().__getitem__(other)
        if isprimitive(value=res):
            return PrimitiveFactory.generate_primitive(value=res)
        else:
            # we can have torch.Tensor and other types
            return res

    def __iter__(self, max_len: Optional[int] = None) -> Iterator:
        return Iterator(super().__iter__(), max_len=max_len)

    def __len__(self) -> SyPrimitiveRet:
        res = super().__len__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = super().__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __reversed__(self) -> Any:
        # returns <class 'odict_iterator'>
        return super().__reversed__()

    def __setitem__(self, key: Any, value: Any) -> None:
        res = super().__setitem__(key, value)
        return PrimitiveFactory.generate_primitive(value=res)

    def clear(self) -> None:
        res = super().clear()
        return PrimitiveFactory.generate_primitive(value=res)

    def copy(self) -> SyPrimitiveRet:
        res = super().copy()
        return PrimitiveFactory.generate_primitive(value=res)

    @classmethod
    def FromKeys(cls, iterable: Any, value: Any = None) -> SyPrimitiveRet:
        res = cls(PyOrderedDict.fromkeys(iterable, value))
        return PrimitiveFactory.generate_primitive(value=res)

    def fromkeys(  # type: ignore
        self, iterable: Any, value: Optional[object] = None
    ) -> SyPrimitiveRet:
        res = super().fromkeys(iterable, value)
        return PrimitiveFactory.generate_primitive(value=res)

    def dict_get(self, other: Any) -> Any:
        res = super().get(other)
        if isprimitive(value=res):
            return PrimitiveFactory.generate_primitive(value=res)
        else:
            # we can have torch.Tensor and other types
            return res

    def items(self, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        return Iterator(ItemsView(self), max_len=max_len)

    def keys(self, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        return Iterator(KeysView(self), max_len=max_len)

    def move_to_end(self, other: Any, last: Any = True) -> Any:
        res = super().move_to_end(other, last)
        return PrimitiveFactory.generate_primitive(value=res)

    def pop(self, *args: Any, **kwargs: Any) -> SyPrimitiveRet:
        res = super().pop(*args, **kwargs)
        return PrimitiveFactory.generate_primitive(value=res)

    def popitem(self, last: Any = True) -> SyPrimitiveRet:
        res = super().popitem(last)
        return PrimitiveFactory.generate_primitive(value=res)

    def setdefault(self, key: Any, default: Optional[object] = None) -> SyPrimitiveRet:
        res = super().setdefault(key, default)
        return PrimitiveFactory.generate_primitive(value=res)

    def update(self, *args, **kwds: Any) -> SyPrimitiveRet:  # type: ignore
        res = super().update(*args, **kwds)
        return PrimitiveFactory.generate_primitive(value=res)

    def values(self, *args: Any, max_len: Optional[int] = None) -> Iterator:  # type: ignore
        # this is what the super type does and there is a test in dict_test.py
        # test_values which checks for this so we could disable the test or
        # keep this workaround
        if len(args) > 0:
            traceback_and_raise(
                TypeError("values() takes 1 positional argument but 2 were given")
            )
        return Iterator(ValuesView(self), max_len=max_len)

    def _object2proto(self) -> OrderedDict_PB:
        id_ = sy.serialize(obj=self.id)
        # serialize to bytes so that we can avoid using StorableObject
        # otherwise we get recursion where the permissions of StorableObject
        # themselves utilise Dict
        keys = [
            sy.serialize(obj=downcast(value=element), to_bytes=True)
            for element in self.keys()
        ]
        # serialize to bytes so that we can avoid using StorableObject
        # otherwise we get recursion where the permissions of StorableObject
        # themselves utilise Dict
        values = [
            sy.serialize(obj=downcast(value=element), to_bytes=True)
            for element in self.values()
        ]
        return OrderedDict_PB(id=id_, keys=keys, values=values)

    @staticmethod
    def _proto2object(proto: OrderedDict_PB) -> "OrderedDict":
        id_: UID = sy.deserialize(blob=proto.id)
        # deserialize from bytes so that we can avoid using StorableObject
        # otherwise we get recursion where the permissions of StorableObject
        # themselves utilise OrderedDict
        values = [
            sy.deserialize(blob=upcast(value=element), from_bytes=True)
            for element in proto.values
        ]
        # deserialize from bytes so that we can avoid using StorableObject
        # otherwise we get recursion where the permissions of StorableObject
        # themselves utilise OrderedDict
        keys = [
            sy.deserialize(blob=upcast(value=element), from_bytes=True)
            for element in proto.keys
        ]
        new_dict = OrderedDict(dict(zip(keys, values)))
        new_dict._id = id_
        return new_dict

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return OrderedDict_PB

    def upcast(self) -> PyOrderedDict:
        # recursively upcast
        return OrderedDict((k, upcast(v)) for k, v in self.items())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/save_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.store import (
    store_object_pb2 as proto_dot_core_dot_store_dot_store__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n/proto/core/node/common/action/save_object.proto\x12\x1csyft.core.node.common.action\x1a#proto/core/store/store_object.proto\x1a\x1bproto/core/io/address.proto"h\n\x10SaveObjectAction\x12,\n\x03obj\x18\x01 \x01(\x0b\x32\x1f.syft.core.store.StorableObject\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_SAVEOBJECTACTION = DESCRIPTOR.message_types_by_name["SaveObjectAction"]
SaveObjectAction = _reflection.GeneratedProtocolMessageType(
    "SaveObjectAction",
    (_message.Message,),
    {
        "DESCRIPTOR": _SAVEOBJECTACTION,
        "__module__": "proto.core.node.common.action.save_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.SaveObjectAction)
    },
)
_sym_db.RegisterMessage(SaveObjectAction)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SAVEOBJECTACTION._serialized_start = 147
    _SAVEOBJECTACTION._serialized_end = 251
# @@protoc_insertion_point(module_scope)



# future
from __future__ import annotations

# stdlib
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

# third party
import jwt
from nacl.signing import VerifyKey
from pydantic import BaseSettings
from typing_extensions import final

# relative
from ......logger import error
from .....common.serde.serializable import serializable
from ....abstract.node_service_interface import NodeServiceInterface
from ...exceptions import InvalidCredentialsError
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


@serializable(recursive_serde=True)
@final
class UserLoginMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class UserLoginReplyMessage(GenericPayloadReplyMessage):
    ...


# TODO: Move create_access_token from grid to syft
def create_access_token(
    settings: BaseSettings,
    subject: Union[str, Any],
    expires_delta: Optional[timedelta] = None,
    ALGORITHM: str = "HS256",
) -> str:
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(
            minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
        )
    to_encode = {"exp": expire, "sub": str(subject)}
    encoded_jwt = jwt.encode(to_encode, settings.SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


@serializable(recursive_serde=True)
@final
class UserLoginMessageWithReply(GenericPayloadMessageWithReply):
    message_type = UserLoginMessage
    message_reply_type = UserLoginReplyMessage

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            email = str(self.kwargs["email"])
            password = str(self.kwargs["password"])
            response: Dict[str, Any] = {}
            response["status"] = "error"
            try:
                node.users.login(email=email, password=password)  # type: ignore
            except InvalidCredentialsError as e:
                error(f"Invalid credentials during login for {email}. {e}")
                response["reason"] = "Incorrect email or password"
                return response

            user = node.users.first(email=email)  # type: ignore

            # settings = node.settings
            # access_token_expires = timedelta(
            #     minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
            # )
            # access_token = create_access_token(
            #     settings=settings, subject=user.id, expires_delta=access_token_expires
            # )
            # metadata = (
            #     sy.serialize(node.get_metadata_for_client())
            #     .SerializeToString()
            #     .decode("ISO-8859-1")
            # )

            data = {
                # "access_token": access_token,
                # "token_type": "bearer",
                # "metadata": metadata,
                "key": user.private_key,
            }

            response["status"] = "ok"
            response["data"] = data

            return response
        except Exception as e:
            error(f"Failed to login with {email}. {e}")
            response["status"] = "error"
            return response



# stdlib
from typing import Any
from typing import Callable as CallableT
from typing import Optional
from typing import Tuple
from typing import Union

# relative
from .. import ast
from ..logger import traceback_and_raise


class DynamicObject(ast.attribute.Attribute):
    """Creates a property in the AST for a dynamically set object in the AST. This family of
    objects aren't present on the object before creating them (compared to properties or slots).
    """

    def __init__(
        self,
        path_and_name: str,
        parent: ast.attribute.Attribute,
        object_ref: Optional[Any] = None,
        return_type_name: Optional[str] = None,
        client: Optional[Any] = None,
    ):
        super().__init__(
            path_and_name=path_and_name,
            parent=parent,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
        )

    def __call__(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Optional[Union[Any, CallableT]]:
        traceback_and_raise(ValueError("Dynamic object should never be called."))



# stdlib
from uuid import UUID

# third party
import numpy as np

# relative
from .....core.common.serde.serializable import serializable
from ..tensor import AutogradTensor
from .op import Op


@serializable(recursive_serde=True)
class SumOp(Op):
    """Sum operation across a dimension"""

    __attr_allowlist__ = ["x", "axis", "dim_at_axis", "backward_shape"]

    def forward(self, x: AutogradTensor, axis: int) -> AutogradTensor:  # type: ignore
        self.x = x
        self.axis = axis
        if axis is not None:
            # obj.sum() can be called without dims
            self.dim_at_axis = self.x.shape[self.axis]
        else:
            self.dim_at_axis = None  # type: ignore
        self.backward_shape = self.x.shape

        result = x.child.sum(axis=axis)

        if result.shape == ():
            result = result.reshape(1)

        return AutogradTensor(result, requires_grad=x.requires_grad)

    def _backward(self, grad: np.ndarray, backprop_id: UUID) -> None:

        if self.x.requires_grad:

            if self.axis is not None:

                grad = np.expand_dims(grad, self.axis)
                grad = grad.repeat(self.dim_at_axis, axis=self.axis)

            else:
                n_times = np.prod(self.backward_shape)  # type: ignore
                grad = grad.repeat(n_times, axis=0).reshape(self.backward_shape)

            self.x.add_grad(grad)

            if self.x.grad_fn:
                self.x.backward(backprop_id=backprop_id)



# third party
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class SetupConfig(Base):
    __tablename__ = "setup"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    domain_name = Column(String(255), default="")
    description = Column(String(255), default="")
    contact = Column(String(255), default="")
    daa = Column(Boolean(), default=False)
    node_id = Column(String(32), default="")
    daa_document = Column(Integer, ForeignKey("daa_pdf.id"))
    tags = Column(String(255), default="[]")
    deployed_on = Column(DateTime())

    def __str__(self) -> str:
        return f"<Domain Name: {self.domain_name}>"


def create_setup(id: int, domain_name: str, node_id: str) -> SetupConfig:
    return SetupConfig(id=id, domain_name=domain_name, node_id=node_id)



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# syft absolute
import syft as sy

# relative
from ......proto.core.node.common.service.get_repr_service_pb2 import (
    GetReprMessage as GetReprMessage_PB,
)
from ......proto.core.node.common.service.get_repr_service_pb2 import (
    GetReprReplyMessage as GetReprReplyMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class GetReprMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        id_at_location: UID,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.id_at_location = id_at_location

    def _object2proto(self) -> GetReprMessage_PB:
        return GetReprMessage_PB(
            id_at_location=sy.serialize(self.id_at_location),
            msg_id=sy.serialize(self.id),
            address=sy.serialize(self.address),
            reply_to=sy.serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: GetReprMessage_PB) -> "GetReprMessage":
        return GetReprMessage(
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
            reply_to=sy.deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetReprMessage_PB


@serializable()
class GetReprReplyMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        repr: str,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.repr = repr

    def _object2proto(self) -> GetReprReplyMessage_PB:
        return GetReprReplyMessage_PB(
            repr=self.repr,
            msg_id=sy.serialize(self.id),
            address=sy.serialize(self.address),
        )

    @staticmethod
    def _proto2object(proto: GetReprReplyMessage_PB) -> "GetReprReplyMessage":
        return GetReprReplyMessage(
            repr=proto.repr,
            msg_id=sy.deserialize(blob=proto.msg_id),
            address=sy.deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetReprReplyMessage_PB



# stdlib
from datetime import datetime

# third party
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base

states = {"creating": 0, "failed": 1, "success": 2, "destroyed": 3}


class Environment(Base):
    __tablename__ = "environment"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    state = Column(Integer)
    provider = Column(String(255))
    region = Column(String(255))
    instance_type = Column(String(255))
    address = Column(String(255), default="0.0.0.0")  # nosec
    syft_address = Column(String(255), default="")  # TODO
    created_at = Column(DateTime, default=datetime.now())
    destroyed_at = Column(DateTime, default=datetime.now())

    def __str__(self) -> str:
        return (
            f"<Group id: {self.id}, state: {self.state}, address: {self.address}, syft_address: "
            f"{self.syft_address}, provider: {self.provider}, region: {self.region}, instance_type: "
            f"{self.instance_type}, created_at: {self.created_at}, destroyed_at: {self.destroyed_at}>"
        )



# # stdlib
# from typing import Callable
# from typing import Dict
# from typing import List
# from typing import Type
# from typing import Union

# # third party
# from nacl.encoding import HexEncoder
# from nacl.signing import SigningKey
# from nacl.signing import VerifyKey
# import torch as th

# # relative
# from .....common.group import VerifyAll
# from .....common.message import ImmediateSyftMessageWithReply
# from .....common.message import SignedImmediateSyftMessageWithoutReply
# from .....common.uid import UID
# from .....store.storeable_object import StorableObject
# from ....abstract.node import AbstractNode
# from ...action.save_object_action import SaveObjectAction
# from ..auth import service_auth
# from ..node_service import ImmediateNodeServiceWithReply
# from .tensor_manager_messages import CreateTensorMessage
# from .tensor_manager_messages import CreateTensorResponse
# from .tensor_manager_messages import DeleteTensorMessage
# from .tensor_manager_messages import DeleteTensorResponse
# from .tensor_manager_messages import GetTensorMessage
# from .tensor_manager_messages import GetTensorResponse
# from .tensor_manager_messages import GetTensorsMessage
# from .tensor_manager_messages import GetTensorsResponse
# from .tensor_manager_messages import UpdateTensorMessage
# from .tensor_manager_messages import UpdateTensorResponse


# def create_tensor_msg(
#     msg: CreateTensorMessage,
#     node: AbstractNode,
# ) -> CreateTensorResponse:
#     try:
#         payload = msg.content

#         new_tensor = th.tensor(payload["tensor"])
#         new_tensor.tag(*payload.get("tags", []))
#         new_tensor.describe(payload.get("description", ""))

#         id_at_location = UID()

#         # Step 2: create message which contains object to send
#         storable = StorableObject(
#             id=id_at_location,
#             data=new_tensor,
#             tags=new_tensor.tags,
#             description=new_tensor.description,
#             search_permissions={VerifyAll(): None}
#             if payload.get("searchable", False)
#             else {},
#         )

#         obj_msg = SaveObjectAction(obj=storable, address=node.address)

#         signed_message: SignedImmediateSyftMessageWithoutReply = obj_msg.sign(
#             signing_key=SigningKey(
#                 payload["internal_key"].encode("utf-8"), encoder=HexEncoder
#             )
#         )

#         node.recv_immediate_msg_without_reply(msg=signed_message)

#         return CreateTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={
#                 "msg": "Tensor created succesfully!",
#                 "tensor_id": str(id_at_location.value),
#             },
#         )
#     except Exception as e:
#         return CreateTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"error": str(e)},
#         )


# def update_tensor_msg(
#     msg: UpdateTensorMessage,
#     node: AbstractNode,
# ) -> UpdateTensorResponse:
#     try:
#         payload = msg.content

#         new_tensor = th.tensor(payload["tensor"])
#         new_tensor.tag(*payload.get("tags", []))
#         new_tensor.describe(payload.get("description", ""))

#         key = UID.from_string(value=payload["tensor_id"])

#         # Step 2: create message which contains object to send
#         storable = StorableObject(
#             id=key,
#             data=new_tensor,
#             tags=new_tensor.tags,
#             description=new_tensor.description,
#             search_permissions={VerifyAll(): None}
#             if payload.get("searchable", False)
#             else {},
#         )

#         obj_msg = SaveObjectAction(obj=storable, address=node.address)

#         signed_message: SignedImmediateSyftMessageWithoutReply = obj_msg.sign(
#             signing_key=SigningKey(
#                 payload["internal_key"].encode("utf-8"), encoder=HexEncoder
#             )
#         )

#         node.recv_immediate_msg_without_reply(msg=signed_message)

#         return UpdateTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"msg": "Tensor modified successfully!"},
#         )
#     except Exception as e:
#         return UpdateTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"error": str(e)},
#         )


# def get_tensor_msg(
#     msg: GetTensorMessage,
#     node: AbstractNode,
# ) -> GetTensorResponse:
#     try:
#         payload = msg.content

#         # Retrieve the dataset from node.store
#         key = UID.from_string(value=payload["tensor_id"])
#         tensor = node.store.get(key)
#         return GetTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={
#                 "tensor": {
#                     "id": payload["tensor_id"],
#                     "tags": tensor.tags,
#                     "description": tensor.description,
#                 }
#             },
#         )
#     except Exception as e:
#         return GetTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"error": str(e)},
#         )


# def get_tensors_msg(
#     msg: GetTensorsMessage,
#     node: AbstractNode,
# ) -> GetTensorsResponse:
#     try:
#         tensors = node.store.get_objects_of_type(obj_type=th.Tensor)

#         result = []

#         for tensor in tensors:
#             result.append(
#                 {
#                     "id": str(tensor.id.value),
#                     "tags": tensor.tags,
#                     "description": tensor.description,
#                 }
#             )
#         return GetTensorsResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"tensors": result},
#         )
#     except Exception as e:
#         return GetTensorsResponse(
#             address=msg.reply_to, content={"error": str(e)}, status_code=200
#         )


# def del_tensor_msg(
#     msg: DeleteTensorMessage,
#     node: AbstractNode,
# ) -> DeleteTensorResponse:
#     try:
#         payload = msg.content

#         # Retrieve the dataset from node.store
#         key = UID.from_string(value=payload["tensor_id"])
#         node.store.delete(key=key)

#         return DeleteTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"msg": "Tensor deleted successfully!"},
#         )
#     except Exception as e:
#         return DeleteTensorResponse(
#             address=msg.reply_to,
#             status_code=200,
#             content={"error": str(e)},
#         )


# class TensorManagerService(ImmediateNodeServiceWithReply):
#     INPUT_TYPE = Union[
#         Type[CreateTensorMessage],
#         Type[UpdateTensorMessage],
#         Type[GetTensorMessage],
#         Type[GetTensorsMessage],
#         Type[DeleteTensorMessage],
#     ]

#     INPUT_MESSAGES = Union[
#         CreateTensorMessage,
#         UpdateTensorMessage,
#         GetTensorMessage,
#         GetTensorsMessage,
#         DeleteTensorMessage,
#     ]

#     OUTPUT_MESSAGES = Union[
#         CreateTensorResponse,
#         UpdateTensorResponse,
#         GetTensorResponse,
#         GetTensorsResponse,
#         DeleteTensorResponse,
#     ]

#     msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
#         CreateTensorMessage: create_tensor_msg,
#         UpdateTensorMessage: update_tensor_msg,
#         GetTensorMessage: get_tensor_msg,
#         GetTensorsMessage: get_tensors_msg,
#         DeleteTensorMessage: del_tensor_msg,
#     }

#     @staticmethod
#     @service_auth(guests_welcome=True)
#     def process(
#         node: AbstractNode,
#         msg: INPUT_MESSAGES,
#         verify_key: VerifyKey,
#     ) -> OUTPUT_MESSAGES:
#         return TensorManagerService.msg_handler_map[type(msg)](msg=msg, node=node)

#     @staticmethod
#     def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
#         return [
#             CreateTensorMessage,
#             UpdateTensorMessage,
#             GetTensorMessage,
#             GetTensorsMessage,
#             DeleteTensorMessage,
#         ]



# stdlib
from typing import Dict

# third party
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

# relative
from . import Base


class AssociationRequest(Base):
    __tablename__ = "association_request"

    id = Column(Integer, primary_key=True, autoincrement=True)
    requested_date = Column(String(255))
    accepted_date = Column(String(255), default="")
    node_name = Column(String(255), default="")
    node_address = Column(String(255), default="")
    name = Column(String(255), default="")
    email = Column(String(255), default="")
    reason = Column(String(255), default="")
    status = Column(String(255), default="")
    source = Column(String(255), default="")
    target = Column(String(255), default="")

    def __str__(self) -> str:
        return (
            f"< Association Request id : {self.id}, Name: {self.name},"
            f" Status: {self.status}, Source: {self.source}, Target: {self.target}"
            f" Date: {self.requested_date}>"
        )

    def get_metadata(self) -> Dict[str, str]:
        return {
            "association_id": str(self.id),
            "requested_date": self.requested_date,
            "name": self.name,
            "email": self.email,
            "reason": self.reason,
            "status": self.status,
            "source": self.source,
            "target": self.target,
            "node_name": self.node_name,
            "node_address": self.node_address,
        }



"""Client AST creation.
"""
# stdlib
from typing import Any
from typing import Optional

# relative
from ...common.uid import UID
from ...node.abstract.node import AbstractNodeClient


def create_client_ast(client: Optional[AbstractNodeClient] = None) -> Any:
    # relative
    from ....ast import add_classes
    from ....ast import add_modules
    from ....ast.globals import Globals
    from .client import Client

    ast = Globals(client)

    modules = [
        "syft",
        "syft.core",
        "syft.core.common",
        "syft.core.common.uid",
        "syft.core.node",
        "syft.core.node.common",
        "syft.core.node.common.client",
    ]

    classes = [
        (
            "syft.core.node.common.client.Client",
            "syft.core.node.common.client.Client",
            Client,
        ),
        (
            "syft.core.common.uid.UID",
            "syft.core.common.uid.UID",
            UID,
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Optional

# third party
from nacl.signing import VerifyKey
import requests
from typing_extensions import final

# relative
from ......grid import GridURL
from .....common.serde.serializable import serializable
from ....abstract.node import AbstractNode
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


def grid_url_from_kwargs(kwargs: Dict[str, Any]) -> GridURL:
    try:
        if "host_or_ip" in kwargs:
            # old way to send these messages was with host_or_ip
            return GridURL.from_url(str(kwargs["host_or_ip"]))
        elif "grid_url" in kwargs:
            # new way is with grid_url
            return kwargs["grid_url"]
        else:
            raise Exception("kwargs missing host_or_ip or grid_url")
    except Exception as e:
        print(f"Failed to get grid_url from kwargs: {kwargs}. {e}")
        raise e


@serializable(recursive_serde=True)
@final
class PingMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class PingReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class PingMessageWithReply(GenericPayloadMessageWithReply):
    message_type = PingMessage
    message_reply_type = PingReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        try:
            grid_url = grid_url_from_kwargs(self.kwargs)
            res = requests.get(str(grid_url.with_path("/status")), timeout=0.25)
            return {
                "grid_url": str(grid_url),
                "result": "ping succeeded",
                "status_code": res.status_code,
            }
        except Exception:
            print("Failed to run ping", self.kwargs)
            return {
                "grid_url": str(grid_url),
                "result": "ping failed",
                "status_code": 200,
            }



# future
from __future__ import annotations

# stdlib
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from nacl.signing import VerifyKey
import numpy as np

# relative
from ....core.adp.entity import Entity
from ....core.adp.entity_list import EntityList
from ....lib.numpy.array import arrow_deserialize as numpy_deserialize
from ....lib.numpy.array import arrow_serialize as numpy_serialize
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.capnp import CapnpModule
from ...common.serde.capnp import chunk_bytes
from ...common.serde.capnp import combine_bytes
from ...common.serde.capnp import get_capnp_schema
from ...common.serde.capnp import serde_magic_header
from ...common.serde.serializable import serializable
from ...common.uid import UID
from ...pointer.pointer import Pointer
from ..ancestors import AutogradTensorAncestor
from ..lazy_repeat_array import lazyrepeatarray
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import SupportedChainType  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore
from .adp_tensor import ADPTensor
from .initial_gamma import InitialGammaTensor
from .initial_gamma import IntermediateGammaTensor
from .single_entity_phi import SingleEntityPhiTensor


@serializable(recursive_serde=True)
class NDimEntityPhiTensorPointer(Pointer):
    __name__ = "NDimEntityPhiTensorPointer"
    __module__ = "syft.core.tensor.autodp.ndim_entity_phi"
    __attr_allowlist__ = [
        # default pointer attrs
        "points_to_object_with_path",
        "pointer_name",
        "id_at_location",
        "location",
        "tags",
        "description",
        "object_type",
        "attribute_name",
        "public_shape",
        "_exhausted",
        "gc_enabled",
        "is_enum",
        # ndim attrs
        "entities",
        "min_vals",
        "max_vals",
        "client",
        "public_dtype",
    ]

    # TODO :should create serialization for Entity List

    def __init__(
        self,
        entities: EntityList,
        min_vals: np.typing.ArrayLike,
        max_vals: np.typing.ArrayLike,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[np.dtype] = None,
    ):
        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.min_vals = min_vals
        self.max_vals = max_vals
        self.entities = entities
        self.public_shape = public_shape
        self.public_dtype = public_dtype


@serializable(capnp_bytes=True)
class NDimEntityPhiTensor(PassthroughTensor, AutogradTensorAncestor, ADPTensor):
    PointerClassOverride = NDimEntityPhiTensorPointer
    # __attr_allowlist__ = ["child", "min_vals", "max_vals", "entities"]
    __slots__ = (
        "child",
        "min_vals",
        "max_vals",
        "entities",
    )

    def __init__(
        self,
        child: Sequence,
        entities: Union[List[Entity], EntityList],
        min_vals: np.ndarray,
        max_vals: np.ndarray,
        row_type: SingleEntityPhiTensor = SingleEntityPhiTensor,  # type: ignore
    ) -> None:

        # child = the actual private data
        super().__init__(child)

        # lazyrepeatarray matching the shape of child
        if not isinstance(min_vals, lazyrepeatarray):
            min_vals = lazyrepeatarray(data=min_vals, shape=child.shape)  # type: ignore
        if not isinstance(max_vals, lazyrepeatarray):
            max_vals = lazyrepeatarray(data=max_vals, shape=child.shape)  # type: ignore
        self.min_vals = min_vals
        self.max_vals = max_vals

        if not isinstance(entities, EntityList):
            entities = EntityList.from_objs(entities)

        self.entities = entities

    @property
    def proxy_public_kwargs(self) -> Dict[str, Any]:
        return {
            "min_vals": self.min_vals,
            "max_vals": self.max_vals,
            "entities": self.entities,
        }

    @staticmethod
    def from_rows(rows: Sequence) -> NDimEntityPhiTensor:
        if len(rows) < 1 or not isinstance(rows[0], SingleEntityPhiTensor):
            raise Exception(
                "NDimEntityPhiTensor.from_rows requires a list of SingleEntityPhiTensors"
            )

        # create lazyrepeatarrays of the first element
        first_row = rows[0]
        min_vals = lazyrepeatarray(
            data=first_row.min_vals,
            shape=tuple([len(rows)] + list(first_row.min_vals.shape)),
        )
        max_vals = lazyrepeatarray(
            data=first_row.max_vals,
            shape=tuple([len(rows)] + list(first_row.max_vals.shape)),
        )

        # collect entities and children into numpy arrays
        entity_list = []
        child_list = []
        for row in rows:
            entity_list.append(row.entity)
            child_list.append(row.child)
        entities = EntityList.from_objs(entities=entity_list)
        child = np.stack(child_list)

        # use new constructor
        return NDimEntityPhiTensor(
            child=child,
            min_vals=min_vals,
            max_vals=max_vals,
            entities=entities,
        )

    def init_pointer(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> NDimEntityPhiTensorPointer:
        return NDimEntityPhiTensorPointer(
            # Arguments specifically for SEPhiTensor
            entities=self.entities,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            # Arguments required for a Pointer to work
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

    @property
    def gamma(self) -> InitialGammaTensor:
        """Property to cast this tensor into a GammaTensor"""
        return self.create_gamma()

    def copy(self, order: Optional[str] = "K") -> NDimEntityPhiTensor:
        """Return copy of the given object"""

        return NDimEntityPhiTensor(
            child=self.child.copy(order=order),
            min_vals=self.min_vals.copy(order=order),
            max_vals=self.max_vals.copy(order=order),
            entities=self.entities.copy(order=order),
        )

    def all(self) -> bool:
        return self.child.all()

    def any(self) -> bool:
        return self.child.any()

    def copy_with(self, child: np.ndarray) -> NDimEntityPhiTensor:
        new_tensor = self.copy()
        new_tensor.child = child
        return new_tensor

    def create_gamma(
        self, scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None
    ) -> InitialGammaTensor:
        """Return a new Gamma tensor based on this phi tensor"""

        # if scalar_manager is None:
        #     scalar_manager = self.scalar_manager

        # Gamma expects an entity for each scalar
        # entities = np.array([self.entity] * np.array(self.child.shape).prod()).reshape(
        #     self.shape
        # )

        # TODO: update InitialGammaTensor to handle EntityList
        return InitialGammaTensor(
            values=self.child,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            entities=self.entities,
            # scalar_manager=scalar_manager,
        )

    def publish(
        self, acc: Any, sigma: float, user_key: VerifyKey
    ) -> AcceptableSimpleType:
        print("PUBLISHING TO GAMMA:")
        print(self.child)
        return self.gamma.publish(acc=acc, sigma=sigma, user_key=user_key)

    @property
    def value(self) -> np.ndarray:
        return self.child

    def astype(self, np_type: np.dtype) -> NDimEntityPhiTensor:
        return self.__class__(
            child=self.child.astype(np_type),
            entities=self.entities,
            min_vals=self.min_vals.astype(np_type),
            max_vals=self.max_vals.astype(np_type),
            # scalar_manager=self.scalar_manager,
        )

    @property
    def shape(self) -> Tuple[Any, ...]:
        return self.child.shape

    def __repr__(self) -> str:
        """Pretty print some information, optimized for Jupyter notebook viewing."""
        return (
            f"{self.__class__.__name__}(child={self.child.shape}, "
            + f"min_vals={self.min_vals}, max_vals={self.max_vals})"
        )

    def __eq__(self, other: Any) -> Union[NDimEntityPhiTensor, IntermediateGammaTensor]:
        # TODO: what about entities and min / max values?
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            gamma_output = False
            if is_acceptable_simple_type(other):
                result = self.child == other
            else:
                # check entities match, if they dont gamma_output = True
                #
                result = self.child == other.child
                if isinstance(result, InitialGammaTensor):
                    gamma_output = True
            if not gamma_output:
                # min_vals=self.min_vals * 0.0,
                # max_vals=self.max_vals * 0.0 + 1.0,
                return self.copy_with(child=result)
            else:
                return self.copy_with(child=result).gamma
        else:
            raise Exception(
                "Tensor dims do not match for __eq__: "
                + f"{len(self.child)} != {len(other.child)}"
            )

    def __add__(
        self, other: SupportedChainType
    ) -> Union[NDimEntityPhiTensor, IntermediateGammaTensor]:

        # if the tensor being added is also private
        if isinstance(other, NDimEntityPhiTensor):
            if self.entities != other.entities:
                return self.gamma + other.gamma

            return NDimEntityPhiTensor(
                child=self.child + other.child,
                min_vals=self.min_vals + other.min_vals,
                max_vals=self.max_vals + other.max_vals,
                entities=self.entities,
                # scalar_manager=self.scalar_manager,
            )

        # if the tensor being added is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):
            return NDimEntityPhiTensor(
                child=self.child + other,
                min_vals=self.min_vals + other,
                max_vals=self.max_vals + other,
                entities=self.entities,
                # scalar_manager=self.scalar_manager,
            )

        elif isinstance(other, IntermediateGammaTensor):
            return self.gamma + other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="ndept.capnp")

        rows, rows_size = numpy_serialize(self.child, get_bytes=True)
        min_vals, min_vals_size = numpy_serialize(self.min_vals.data, get_bytes=True)
        max_vals, max_vals_size = numpy_serialize(self.max_vals.data, get_bytes=True)
        entities_indexed, entities_indexed_size = numpy_serialize(
            self.entities.entities_indexed, get_bytes=True
        )
        one_hot_lookup = self.entities.one_hot_lookup

        ndept_struct: CapnpModule = schema.NDEPT  # type: ignore
        ndept_msg = ndept_struct.new_message()
        metadata_schema = ndept_struct.TensorMetadata
        child_metadata = metadata_schema.new_message()
        min_vals_metadata = metadata_schema.new_message()
        max_vals_metadata = metadata_schema.new_message()
        entities_metadata = metadata_schema.new_message()

        # this is how we dispatch correct deserialization of bytes
        ndept_msg.magicHeader = serde_magic_header(type(self))

        chunk_bytes(rows, "child", ndept_msg)
        child_metadata.dtype = str(self.child.dtype)
        child_metadata.decompressedSize = rows_size
        ndept_msg.childMetadata = child_metadata

        chunk_bytes(min_vals, "minVals", ndept_msg)
        min_vals_metadata.dtype = str(self.min_vals.data.dtype)
        min_vals_metadata.decompressedSize = min_vals_size
        ndept_msg.minValsMetadata = min_vals_metadata

        chunk_bytes(max_vals, "maxVals", ndept_msg)
        max_vals_metadata.dtype = str(self.max_vals.data.dtype)
        max_vals_metadata.decompressedSize = max_vals_size
        ndept_msg.maxValsMetadata = max_vals_metadata

        chunk_bytes(entities_indexed, "entitiesIndexed", ndept_msg)
        entities_metadata.dtype = str(self.entities.entities_indexed.dtype)
        entities_metadata.decompressedSize = entities_indexed_size
        ndept_msg.entitiesIndexedMetadata = entities_metadata

        oneHotLookupList = ndept_msg.init("oneHotLookup", len(one_hot_lookup))
        for i, entity in enumerate(one_hot_lookup):
            oneHotLookupList[i] = (
                entity if not getattr(entity, "name", None) else entity.name  # type: ignore
            )

        # to pack or not to pack?
        # return ndept_msg.to_bytes()
        return ndept_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> NDimEntityPhiTensor:
        schema = get_capnp_schema(schema_file="ndept.capnp")
        ndept_struct: CapnpModule = schema.NDEPT  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        # ndept_msg = ndept_struct.from_bytes(buf, traversal_limit_in_words=2 ** 64 - 1)
        ndept_msg = ndept_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        child_metadata = ndept_msg.childMetadata

        child = numpy_deserialize(
            combine_bytes(ndept_msg.child),
            child_metadata.decompressedSize,
            child_metadata.dtype,
        )

        min_vals_metadata = ndept_msg.minValsMetadata
        min_vals = lazyrepeatarray(
            numpy_deserialize(
                combine_bytes(ndept_msg.minVals),
                min_vals_metadata.decompressedSize,
                min_vals_metadata.dtype,
            ),
            child.shape,
        )

        max_vals_metadata = ndept_msg.maxValsMetadata
        max_vals = lazyrepeatarray(
            numpy_deserialize(
                combine_bytes(ndept_msg.maxVals),
                max_vals_metadata.decompressedSize,
                max_vals_metadata.dtype,
            ),
            child.shape,
        )

        entities_metadata = ndept_msg.entitiesIndexedMetadata
        entities_indexed = numpy_deserialize(
            combine_bytes(ndept_msg.entitiesIndexed),
            entities_metadata.decompressedSize,
            entities_metadata.dtype,
        )
        one_hot_lookup = np.array(ndept_msg.oneHotLookup)

        entity_list = EntityList(one_hot_lookup, entities_indexed)

        return NDimEntityPhiTensor(
            child=child, min_vals=min_vals, max_vals=max_vals, entities=entity_list
        )



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.core.node.domain.service.get_remaining_budget_service_pb2 import (
    GetRemainingBudgetMessage as GetRemainingBudgetMessage_PB,
)
from ......proto.core.node.domain.service.get_remaining_budget_service_pb2 import (
    GetRemainingBudgetReplyMessage as GetRemainingBudgetReplyMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class GetRemainingBudgetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetRemainingBudgetMessage_PB:
        return GetRemainingBudgetMessage_PB(
            # budget=self.budget,
            # id_at_location=serialize(self.id_at_location),
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetRemainingBudgetMessage_PB,
    ) -> "GetRemainingBudgetMessage":
        return GetRemainingBudgetMessage(
            # budget=proto.budget,
            # id_at_location=_deserialize(blob=proto.id_at_location),
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetRemainingBudgetMessage_PB


@serializable()
class GetRemainingBudgetReplyMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        budget: float,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.budget = budget

    def _object2proto(self) -> GetRemainingBudgetReplyMessage_PB:
        return GetRemainingBudgetReplyMessage_PB(
            budget=self.budget,
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(
        proto: GetRemainingBudgetReplyMessage_PB,
    ) -> "GetRemainingBudgetReplyMessage":
        return GetRemainingBudgetReplyMessage(
            budget=proto.budget,
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetRemainingBudgetReplyMessage_PB



# stdlib
from datetime import datetime
from json import dumps
from json import loads
from typing import Callable
from typing import Dict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey

# relative
from .....common import UID
from .....common.message import ImmediateSyftMessageWithReply
from .....io.location import SpecificLocation
from ....domain.domain_interface import DomainInterface
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...exceptions import OwnerAlreadyExistsError
from ...node_table.utils import model_to_json
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .node_setup_messages import CreateInitialSetUpMessage
from .node_setup_messages import GetSetUpMessage
from .node_setup_messages import GetSetUpResponse
from .node_setup_messages import UpdateSetupMessage
from .node_setup_messages import UpdateSetupResponse


def set_node_uid(node: DomainInterface) -> None:
    try:
        setup = node.setup.first()
    except Exception as e:
        print("Missing Setup Table entry", e)

    try:
        node_id = UID.from_string(setup.node_id)
    except Exception as e:
        print(f"Invalid Node UID in Setup Table. {setup.node_id}")
        raise e

    location = SpecificLocation(name=setup.domain_name, id=node_id)
    # TODO: Fix with proper isinstance when the class will import
    if type(node).__name__ == "Domain":
        node.domain = location
    elif type(node).__name__ == "Network":
        node.network = location
    print(f"Finished setting Node UID. {location}")


def create_initial_setup(
    msg: CreateInitialSetUpMessage, node: DomainInterface, verify_key: VerifyKey
) -> SuccessResponseMessage:

    print("Performing initial setup...")

    # 1 - Should not run if Node has an owner
    if len(node.users):
        set_node_uid(node=node)  # make sure the node always has the same UID
        raise OwnerAlreadyExistsError

    # 2 - Check if email/password/node_name fields are empty
    _mandatory_request_fields = msg.email and msg.password and msg.domain_name
    if not _mandatory_request_fields:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (email/password/domain_name)!"
        )

    # 3 - Change Node Name
    node.name = msg.domain_name

    # 4 - Create Admin User
    _node_private_key = node.signing_key.encode(encoder=HexEncoder).decode("utf-8")  # type: ignore

    _verify_key = node.signing_key.verify_key.encode(encoder=HexEncoder).decode("utf-8")  # type: ignore

    _admin_role = node.roles.owner_role

    _ = node.users.signup(
        name=msg.name,
        email=msg.email,
        password=msg.password,
        role=_admin_role.id,
        budget=msg.budget,
        private_key=_node_private_key,
        verify_key=_verify_key,
    )

    # 5 - Save Node SetUp Configs
    try:
        node_id = node.target_id.id
        node.setup.register(
            domain_name=msg.domain_name,
            node_id=node_id.no_dash,
            deployed_on=datetime.now(),
        )
    except Exception as e:
        print("Failed to save setup to database", e)

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Running initial setup!",
    )


def get_setup(
    msg: GetSetUpMessage, node: DomainInterface, verify_key: VerifyKey
) -> GetSetUpResponse:

    _setup = model_to_json(node.setup.first(domain_name=node.name))
    _setup["tags"] = loads(_setup["tags"])
    if node.network:
        _setup["domains"] = len(node.node.all())
    return GetSetUpResponse(
        address=msg.reply_to,
        content=_setup,
    )


def update_settings(
    msg: UpdateSetupMessage, node: DomainInterface, verify_key: VerifyKey
) -> UpdateSetupResponse:
    if node.users.role(verify_key=verify_key).id == node.roles.owner_role.id:
        if msg.domain_name:
            node.name = msg.domain_name

        node.setup.update(
            domain_name=node.name,
            description=msg.description,
            daa=msg.daa,
            contact=msg.contact,
            daa_document=msg.daa_document,
            tags=dumps(msg.tags),
        )
    else:
        raise AuthorizationError("You're not allowed to get setup configs!")

    return UpdateSetupResponse(
        address=msg.reply_to,
        content={"message": "Node settings have been updated successfully!"},
    )


class NodeSetupService(ImmediateNodeServiceWithReply):

    msg_handler_map: Dict[type, Callable] = {
        CreateInitialSetUpMessage: create_initial_setup,
        GetSetUpMessage: get_setup,
        UpdateSetupMessage: update_settings,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: Union[
            CreateInitialSetUpMessage,
            GetSetUpMessage,
            UpdateSetupMessage,
        ],
        verify_key: VerifyKey,
    ) -> Union[SuccessResponseMessage, GetSetUpResponse, UpdateSetupMessage]:
        return NodeSetupService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateInitialSetUpMessage,
            GetSetUpMessage,
            UpdateSetupMessage,
        ]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/domain/service/request_answer_response.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n<proto/core/node/domain/service/request_answer_response.proto\x12\x1dsyft.core.node.domain.service\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"z\n\x15RequestAnswerResponse\x12\x0e\n\x06status\x18\x01 \x01(\x05\x12)\n\nrequest_id\x18\x02 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_REQUESTANSWERRESPONSE = DESCRIPTOR.message_types_by_name["RequestAnswerResponse"]
RequestAnswerResponse = _reflection.GeneratedProtocolMessageType(
    "RequestAnswerResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REQUESTANSWERRESPONSE,
        "__module__": "proto.core.node.domain.service.request_answer_response_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.domain.service.RequestAnswerResponse)
    },
)
_sym_db.RegisterMessage(RequestAnswerResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _REQUESTANSWERRESPONSE._serialized_start = 163
    _REQUESTANSWERRESPONSE._serialized_end = 285
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import cast

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ..... import serialize
from ....common.uid import UID
from ..node_table.bin_obj_dataset import BinObjDataset
from ..node_table.dataset import Dataset
from .database_manager import DatabaseManager


class DatasetManager(DatabaseManager):

    schema = Dataset

    def __init__(self, database: Engine) -> None:
        super().__init__(schema=DatasetManager.schema, db=database)

    def register(self, **kwargs: Any) -> str:
        """Register e  new object into the database.

        Args:
            parameters : List of object parameters.
        Returns:
            object: Database Object
        """
        tags = list(map(str, kwargs.get("tags", [])))
        manifest = str(kwargs.get("manifest", ""))
        name = str(kwargs.get("name", ""))
        description = str(kwargs.get("description", ""))

        print("All the dataset arguments:")
        print(kwargs)
        blob_metadata = {}
        str_metadata = {}
        for key, value in kwargs.items():
            if (
                key != "tags"
                and key != "manifest"
                and key != "description"
                and key != "name"
            ):
                if isinstance(key, str) and isinstance(value, str):
                    str_metadata[key] = value
                else:
                    blob_metadata[str(key)] = cast(
                        bytes, serialize(value, to_bytes=True)
                    ).hex()

        _obj = self._schema(
            id=str(UID().value),
            name=name,
            tags=tags,
            manifest=manifest,
            description=description,
            str_metadata=str_metadata,
            blob_metadata=blob_metadata,
        )
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.add(_obj)
        obj_id = _obj.id
        session_local.commit()
        session_local.close()
        return obj_id

    def add(
        self, name: str, dataset_id: str, obj_id: str, dtype: str, shape: str
    ) -> None:

        obj_dataset_relation = BinObjDataset(
            name=name,
            dataset=dataset_id,
            obj=obj_id,
            dtype=dtype,
            shape=shape,
        )
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        session_local.add(obj_dataset_relation)
        session_local.commit()
        session_local.close()

    def get(self, dataset_id: str) -> Tuple[Dataset, List[bytes]]:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        ds = session_local.query(Dataset).filter_by(id=dataset_id).first()
        objs = list(
            session_local.query(BinObjDataset).filter_by(dataset=dataset_id).all()
        )
        session_local.close()
        return ds, objs

    def set(self, dataset_id: str, metadata: Dict) -> None:
        self.modify({"id": dataset_id}, metadata)
